"""
Create ssh executor system
"""

import logging
import os
import time

import salt.client.ssh.shell
import salt.client.ssh.state
import salt.defaults.exitcodes
import salt.loader
import salt.minion
import salt.roster
import salt.state
import salt.utils.args
import salt.utils.data
import salt.utils.files
import salt.utils.hashutils
import salt.utils.jid
import salt.utils.json
import salt.utils.platform
import salt.utils.state
import salt.utils.thin
from salt.exceptions import SaltInvocationError

__func_alias__ = {"apply_": "apply"}
log = logging.getLogger(__name__)


def _ssh_state(chunks, st_kwargs, kwargs, pillar, test=False):
    """
    Function to run a state with the given chunk via salt-ssh
    """
    file_refs = salt.client.ssh.state.lowstate_file_refs(
        chunks,
        _merge_extra_filerefs(
            kwargs.get("extra_filerefs", ""), __opts__.get("extra_filerefs", "")
        ),
    )
    # Create the tar containing the state pkg and relevant files.
    trans_tar = salt.client.ssh.state.prep_trans_tar(
        __context__["fileclient"],
        chunks,
        file_refs,
        pillar,
        st_kwargs["id_"],
    )
    trans_tar_sum = salt.utils.hashutils.get_hash(trans_tar, __opts__["hash_type"])
    cmd = "state.pkg {}/salt_state.tgz test={} pkg_sum={} hash_type={}".format(
        __opts__["thin_dir"], test, trans_tar_sum, __opts__["hash_type"]
    )
    single = salt.client.ssh.Single(
        __opts__,
        cmd,
        fsclient=__context__["fileclient"],
        minion_opts=__salt__.minion_opts,
        **st_kwargs,
    )
    single.shell.send(trans_tar, "{}/salt_state.tgz".format(__opts__["thin_dir"]))
    stdout, stderr, _ = single.cmd_block()

    # Clean up our tar
    try:
        os.remove(trans_tar)
    except OSError:
        pass

    # Read in the JSON data and return the data structure
    try:
        return salt.utils.data.decode(
            salt.utils.json.loads(stdout, object_hook=salt.utils.data.encode_dict)
        )
    except Exception as e:  # pylint: disable=broad-except
        log.error("JSON Render failed for: %s\n%s", stdout, stderr)
        log.error(str(e))

    # If for some reason the json load fails, return the stdout
    return salt.utils.data.decode(stdout)


def _set_retcode(ret, highstate=None):
    """
    Set the return code based on the data back from the state system
    """

    # Set default retcode to 0
    __context__["retcode"] = salt.defaults.exitcodes.EX_OK

    if isinstance(ret, list):
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return
    if not salt.utils.state.check_result(ret, highstate=highstate):

        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_FAILURE


def _check_pillar(kwargs, pillar=None):
    """
    Check the pillar for errors, refuse to run the state if there are errors
    in the pillar and return the pillar errors
    """
    if kwargs.get("force"):
        return True
    pillar_dict = pillar if pillar is not None else __pillar__.value()
    if "_errors" in pillar_dict:
        return False
    return True


def _wait(jid):
    """
    Wait for all previously started state jobs to finish running
    """
    if jid is None:
        jid = salt.utils.jid.gen_jid(__opts__)
    states = _prior_running_states(jid)
    while states:
        time.sleep(1)
        states = _prior_running_states(jid)


def _merge_extra_filerefs(*args):
    """
    Takes a list of filerefs and returns a merged list
    """
    ret = []
    for arg in args:
        if isinstance(arg, str):
            if arg:
                ret.extend(arg.split(","))
        elif isinstance(arg, list):
            if arg:
                ret.extend(arg)
    return ",".join(ret)


def _cleanup_slsmod_low_data(low_data):
    """
    Set "slsmod" keys to None to make
    low_data JSON serializable
    """
    for i in low_data:
        if "slsmod" in i:
            i["slsmod"] = None


def _cleanup_slsmod_high_data(high_data):
    """
    Set "slsmod" keys to None to make
    high_data JSON serializable
    """
    for i in high_data.values():
        if "stateconf" in i:
            stateconf_data = i["stateconf"][1]
            if "slsmod" in stateconf_data:
                stateconf_data["slsmod"] = None


def _parse_mods(mods):
    """
    Parse modules.
    """
    if isinstance(mods, str):
        mods = [item.strip() for item in mods.split(",") if item.strip()]

    return mods


def sls(mods, saltenv="base", test=None, exclude=None, **kwargs):
    """
    Create the seed file for a state.sls run
    """
    st_kwargs = __salt__.kwargs
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)
    initial_pillar = _get_initial_pillar(opts)
    pillar_override = kwargs.get("pillar")
    with salt.client.ssh.state.SSHHighState(
        opts,
        pillar_override,
        __salt__.value(),
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:
        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += st_.opts["pillar"]["_errors"]
            return err
        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        st_.push_active()
        mods = _parse_mods(mods)
        high_data, errors = st_.render_highstate(
            {saltenv: mods}, context=__context__.value()
        )
        if exclude:
            if isinstance(exclude, str):
                exclude = exclude.split(",")
            if "__exclude__" in high_data:
                high_data["__exclude__"].extend(exclude)
            else:
                high_data["__exclude__"] = exclude
        high_data, ext_errors = st_.state.reconcile_extend(high_data)
        errors += ext_errors
        errors += st_.state.verify_high(high_data)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        high_data, req_in_errors = st_.state.requisite_in(high_data)
        errors += req_in_errors
        high_data = st_.state.apply_exclude(high_data)
        # Verify that the high data is structurally sound
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        # Compile and verify the raw chunks
        chunks = st_.state.compile_high_data(high_data)
        file_refs = salt.client.ssh.state.lowstate_file_refs(
            chunks,
            _merge_extra_filerefs(
                kwargs.get("extra_filerefs", ""), opts.get("extra_filerefs", "")
            ),
        )

        roster = salt.roster.Roster(opts, opts.get("roster", "flat"))
        roster_grains = roster.opts["grains"]

        # Create the tar containing the state pkg and relevant files.
        _cleanup_slsmod_low_data(chunks)
        trans_tar = salt.client.ssh.state.prep_trans_tar(
            __context__["fileclient"],
            chunks,
            file_refs,
            pillar,
            st_kwargs["id_"],
            roster_grains,
        )
        trans_tar_sum = salt.utils.hashutils.get_hash(trans_tar, opts["hash_type"])
        cmd = "state.pkg {}/salt_state.tgz test={} pkg_sum={} hash_type={}".format(
            opts["thin_dir"], test, trans_tar_sum, opts["hash_type"]
        )
        single = salt.client.ssh.Single(
            opts,
            cmd,
            fsclient=__context__["fileclient"],
            minion_opts=__salt__.minion_opts,
            **st_kwargs,
        )
        single.shell.send(trans_tar, "{}/salt_state.tgz".format(opts["thin_dir"]))
        stdout, stderr, _ = single.cmd_block()

        # Clean up our tar
        try:
            os.remove(trans_tar)
        except OSError:
            pass

        # Read in the JSON data and return the data structure
        try:
            return salt.utils.json.loads(stdout)
        except Exception as e:  # pylint: disable=broad-except
            log.error("JSON Render failed for: %s\n%s", stdout, stderr)
            log.error(str(e))

        # If for some reason the json load fails, return the stdout
        return stdout


def running(concurrent=False):
    """
    Return a list of strings that contain state return data if a state function
    is already running. This function is used to prevent multiple state calls
    from being run at the same time.

    CLI Example:

    .. code-block:: bash

        salt '*' state.running
    """
    ret = []
    if concurrent:
        return ret
    active = __salt__["saltutil.is_running"]("state.*")
    for data in active:
        err = (
            'The function "{}" is running as PID {} and was started at {} '
            "with jid {}".format(
                data["fun"],
                data["pid"],
                salt.utils.jid.jid_to_time(data["jid"]),
                data["jid"],
            )
        )
        ret.append(err)
    return ret


def _prior_running_states(jid):
    """
    Return a list of dicts of prior calls to state functions.  This function is
    used to queue state calls so only one is run at a time.
    """

    ret = []
    active = __salt__["saltutil.is_running"]("state.*")
    for data in active:
        try:
            data_jid = int(data["jid"])
        except ValueError:
            continue
        if data_jid < int(jid):
            ret.append(data)
    return ret


def _check_queue(queue, kwargs):
    """
    Utility function to queue the state run if requested
    and to check for conflicts in currently running states
    """
    if queue:
        _wait(kwargs.get("__pub_jid"))
    else:
        conflict = running(concurrent=kwargs.get("concurrent", False))
        if conflict:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return conflict


def _get_initial_pillar(opts):
    return __pillar__.value() if opts["pillarenv"] == __opts__["pillarenv"] else None


def low(data, **kwargs):
    """
    Execute a single low data call
    This function is mostly intended for testing the state system

    CLI Example:

    .. code-block:: bash

        salt '*' state.low '{"state": "pkg", "fun": "installed", "name": "vi"}'
    """
    st_kwargs = __salt__.kwargs
    __opts__["grains"] = __grains__.value()
    chunks = [data]
    with salt.client.ssh.state.SSHHighState(
        __opts__,
        None,
        __salt__.value(),
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=__pillar__.value(),
    ) as st_:
        for chunk in chunks:
            chunk["__id__"] = (
                chunk["name"] if not chunk.get("__id__") else chunk["__id__"]
            )
        err = st_.state.verify_data(data)
        if err:
            return err
        file_refs = salt.client.ssh.state.lowstate_file_refs(
            chunks,
            _merge_extra_filerefs(
                kwargs.get("extra_filerefs", ""), __opts__.get("extra_filerefs", "")
            ),
        )
        roster = salt.roster.Roster(__opts__, __opts__.get("roster", "flat"))
        roster_grains = roster.opts["grains"]

        # Create the tar containing the state pkg and relevant files.
        trans_tar = salt.client.ssh.state.prep_trans_tar(
            __context__["fileclient"],
            chunks,
            file_refs,
            __pillar__.value(),
            st_kwargs["id_"],
            roster_grains,
        )
        trans_tar_sum = salt.utils.hashutils.get_hash(trans_tar, __opts__["hash_type"])
        cmd = "state.pkg {}/salt_state.tgz pkg_sum={} hash_type={}".format(
            __opts__["thin_dir"], trans_tar_sum, __opts__["hash_type"]
        )
        single = salt.client.ssh.Single(
            __opts__,
            cmd,
            fsclient=__context__["fileclient"],
            minion_opts=__salt__.minion_opts,
            **st_kwargs,
        )
        single.shell.send(trans_tar, "{}/salt_state.tgz".format(__opts__["thin_dir"]))
        stdout, stderr, _ = single.cmd_block()

        # Clean up our tar
        try:
            os.remove(trans_tar)
        except OSError:
            pass

        # Read in the JSON data and return the data structure
        try:
            return salt.utils.json.loads(stdout)
        except Exception as e:  # pylint: disable=broad-except
            log.error("JSON Render failed for: %s\n%s", stdout, stderr)
            log.error(str(e))

        # If for some reason the json load fails, return the stdout
        return stdout


def _get_test_value(test=None, **kwargs):
    """
    Determine the correct value for the test flag.
    """
    ret = True
    if test is None:
        if salt.utils.args.test_mode(test=test, **kwargs):
            ret = True
        else:
            ret = __opts__.get("test", None)
    else:
        ret = test
    return ret


def high(data, **kwargs):
    """
    Execute the compound calls stored in a single set of high data
    This function is mostly intended for testing the state system

    CLI Example:

    .. code-block:: bash

        salt '*' state.high '{"vim": {"pkg": ["installed"]}}'
    """
    st_kwargs = __salt__.kwargs
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    pillar_override = kwargs.get("pillar")
    initial_pillar = _get_initial_pillar(opts)
    with salt.client.ssh.state.SSHHighState(
        opts,
        pillar_override,
        __salt__.value(),
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:
        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        st_.push_active()
        chunks = st_.state.compile_high_data(data)
        file_refs = salt.client.ssh.state.lowstate_file_refs(
            chunks,
            _merge_extra_filerefs(
                kwargs.get("extra_filerefs", ""), opts.get("extra_filerefs", "")
            ),
        )

        roster = salt.roster.Roster(opts, opts.get("roster", "flat"))
        roster_grains = roster.opts["grains"]

        # Create the tar containing the state pkg and relevant files.
        _cleanup_slsmod_low_data(chunks)
        trans_tar = salt.client.ssh.state.prep_trans_tar(
            __context__["fileclient"],
            chunks,
            file_refs,
            pillar,
            st_kwargs["id_"],
            roster_grains,
        )
        trans_tar_sum = salt.utils.hashutils.get_hash(trans_tar, opts["hash_type"])
        cmd = "state.pkg {}/salt_state.tgz pkg_sum={} hash_type={}".format(
            opts["thin_dir"], trans_tar_sum, opts["hash_type"]
        )
        single = salt.client.ssh.Single(
            opts,
            cmd,
            fsclient=__context__["fileclient"],
            minion_opts=__salt__.minion_opts,
            **st_kwargs,
        )
        single.shell.send(trans_tar, "{}/salt_state.tgz".format(opts["thin_dir"]))
        stdout, stderr, _ = single.cmd_block()

        # Clean up our tar
        try:
            os.remove(trans_tar)
        except OSError:
            pass

        # Read in the JSON data and return the data structure
        try:
            return salt.utils.json.loads(stdout)
        except Exception as e:  # pylint: disable=broad-except
            log.error("JSON Render failed for: %s\n%s", stdout, stderr)
            log.error(str(e))

        # If for some reason the json load fails, return the stdout
        return stdout


def apply_(mods=None, **kwargs):
    """
    .. versionadded:: 2015.5.3

    Apply states! This function will call highstate or state.sls based on the
    arguments passed in, state.apply is intended to be the main gateway for
    all state executions.

    CLI Example:

    .. code-block:: bash

        salt '*' state.apply
        salt '*' state.apply test
        salt '*' state.apply test,pkgs
    """
    if mods:
        return sls(mods, **kwargs)
    return highstate(**kwargs)


def request(mods=None, **kwargs):
    """
    .. versionadded:: 2017.7.3

    Request that the local admin execute a state run via
    `salt-call state.run_request`
    All arguments match state.apply

    CLI Example:

    .. code-block:: bash

        salt '*' state.request
        salt '*' state.request test
        salt '*' state.request test,pkgs
    """
    kwargs["test"] = True
    ret = apply_(mods, **kwargs)
    notify_path = os.path.join(__opts__["cachedir"], "req_state.p")
    req = check_request()
    req.update(
        {
            kwargs.get("name", "default"): {
                "test_run": ret,
                "mods": mods,
                "kwargs": kwargs,
            }
        }
    )
    with salt.utils.files.set_umask(0o077):
        try:
            if salt.utils.platform.is_windows():
                # Make sure cache file isn't read-only
                __salt__["cmd.run"](f'attrib -R "{notify_path}"')
            with salt.utils.files.fopen(notify_path, "w+b") as fp_:
                salt.payload.dump(req, fp_)
        except OSError:
            log.error(
                "Unable to write state request file %s. Check permission.", notify_path
            )
    return ret


def check_request(name=None):
    """
    .. versionadded:: 2017.7.3

    Return the state request information, if any

    CLI Example:

    .. code-block:: bash

        salt '*' state.check_request
    """
    notify_path = os.path.join(__opts__["cachedir"], "req_state.p")
    if os.path.isfile(notify_path):
        with salt.utils.files.fopen(notify_path, "rb") as fp_:
            # Not sure if this needs to be decoded since it is being returned,
            # and msgpack serialization will encode it to bytes anyway.
            req = salt.payload.load(fp_)
        if name:
            return req[name]
        return req
    return {}


def clear_request(name=None):
    """
    .. versionadded:: 2017.7.3

    Clear out the state execution request without executing it

    CLI Example:

    .. code-block:: bash

        salt '*' state.clear_request
    """
    notify_path = os.path.join(__opts__["cachedir"], "req_state.p")
    if not os.path.isfile(notify_path):
        return True
    if not name:
        try:
            os.remove(notify_path)
        except OSError:
            pass
    else:
        req = check_request()
        if name in req:
            req.pop(name)
        else:
            return False
        with salt.utils.files.set_umask(0o077):
            try:
                if salt.utils.platform.is_windows():
                    # Make sure cache file isn't read-only
                    __salt__["cmd.run"](f'attrib -R "{notify_path}"')
                with salt.utils.files.fopen(notify_path, "w+b") as fp_:
                    salt.payload.dump(req, fp_)
            except OSError:
                log.error(
                    "Unable to write state request file %s. Check permission.",
                    notify_path,
                )
    return True


def run_request(name="default", **kwargs):
    """
    .. versionadded:: 2017.7.3

    Execute the pending state request

    CLI Example:

    .. code-block:: bash

        salt '*' state.run_request
    """
    req = check_request()
    if name not in req:
        return {}
    n_req = req[name]
    if "mods" not in n_req or "kwargs" not in n_req:
        return {}
    req[name]["kwargs"].update(kwargs)
    if "test" in n_req["kwargs"]:
        n_req["kwargs"].pop("test")
    if req:
        ret = apply_(n_req["mods"], **n_req["kwargs"])
        try:
            os.remove(os.path.join(__opts__["cachedir"], "req_state.p"))
        except OSError:
            pass
        return ret
    return {}


def highstate(test=None, **kwargs):
    """
    Retrieve the state data from the salt master for this minion and execute it

    CLI Example:

    .. code-block:: bash

        salt '*' state.highstate

        salt '*' state.highstate exclude=sls_to_exclude
        salt '*' state.highstate exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"
    """
    st_kwargs = __salt__.kwargs
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)
    pillar_override = kwargs.get("pillar")
    initial_pillar = _get_initial_pillar(opts)
    with salt.client.ssh.state.SSHHighState(
        opts,
        pillar_override,
        __salt__.value(),
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:
        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += st_.opts["pillar"]["_errors"]
            return err
        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        st_.push_active()
        chunks = st_.compile_low_chunks(context=__context__.value())
        file_refs = salt.client.ssh.state.lowstate_file_refs(
            chunks,
            _merge_extra_filerefs(
                kwargs.get("extra_filerefs", ""), opts.get("extra_filerefs", "")
            ),
        )
        # Check for errors
        for chunk in chunks:
            if not isinstance(chunk, dict):
                __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
                return chunks

        roster = salt.roster.Roster(opts, opts.get("roster", "flat"))
        roster_grains = roster.opts["grains"]

        # Create the tar containing the state pkg and relevant files.
        _cleanup_slsmod_low_data(chunks)
        trans_tar = salt.client.ssh.state.prep_trans_tar(
            __context__["fileclient"],
            chunks,
            file_refs,
            pillar,
            st_kwargs["id_"],
            roster_grains,
        )
        trans_tar_sum = salt.utils.hashutils.get_hash(trans_tar, opts["hash_type"])
        cmd = "state.pkg {}/salt_state.tgz test={} pkg_sum={} hash_type={}".format(
            opts["thin_dir"], test, trans_tar_sum, opts["hash_type"]
        )
        single = salt.client.ssh.Single(
            opts,
            cmd,
            fsclient=__context__["fileclient"],
            minion_opts=__salt__.minion_opts,
            **st_kwargs,
        )
        single.shell.send(trans_tar, "{}/salt_state.tgz".format(opts["thin_dir"]))
        stdout, stderr, _ = single.cmd_block()

        # Clean up our tar
        try:
            os.remove(trans_tar)
        except OSError:
            pass

        # Read in the JSON data and return the data structure
        try:
            return salt.utils.json.loads(stdout)
        except Exception as e:  # pylint: disable=broad-except
            log.error("JSON Render failed for: %s\n%s", stdout, stderr)
            log.error(str(e))

        # If for some reason the json load fails, return the stdout
        return stdout


def top(topfn, test=None, **kwargs):
    """
    Execute a specific top file instead of the default

    CLI Example:

    .. code-block:: bash

        salt '*' state.top reverse_top.sls
        salt '*' state.top reverse_top.sls exclude=sls_to_exclude
        salt '*' state.top reverse_top.sls exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"
    """
    st_kwargs = __salt__.kwargs
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)
    pillar_override = kwargs.get("pillar")
    initial_pillar = _get_initial_pillar(opts)
    with salt.client.ssh.state.SSHHighState(
        opts,
        pillar_override,
        __salt__.value(),
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:
        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += st_.opts["pillar"]["_errors"]
            return err
        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        st_.opts["state_top"] = os.path.join("salt://", topfn)
        st_.push_active()
        chunks = st_.compile_low_chunks(context=__context__.value())
        # Check for errors
        for chunk in chunks:
            if not isinstance(chunk, dict):
                __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
                return chunks
        file_refs = salt.client.ssh.state.lowstate_file_refs(
            chunks,
            _merge_extra_filerefs(
                kwargs.get("extra_filerefs", ""), opts.get("extra_filerefs", "")
            ),
        )

        roster = salt.roster.Roster(opts, opts.get("roster", "flat"))
        roster_grains = roster.opts["grains"]

        # Create the tar containing the state pkg and relevant files.
        _cleanup_slsmod_low_data(chunks)
        trans_tar = salt.client.ssh.state.prep_trans_tar(
            __context__["fileclient"],
            chunks,
            file_refs,
            pillar,
            st_kwargs["id_"],
            roster_grains,
        )
        trans_tar_sum = salt.utils.hashutils.get_hash(trans_tar, opts["hash_type"])
        cmd = "state.pkg {}/salt_state.tgz test={} pkg_sum={} hash_type={}".format(
            opts["thin_dir"], test, trans_tar_sum, opts["hash_type"]
        )
        single = salt.client.ssh.Single(
            opts,
            cmd,
            fsclient=__context__["fileclient"],
            minion_opts=__salt__.minion_opts,
            **st_kwargs,
        )
        single.shell.send(trans_tar, "{}/salt_state.tgz".format(opts["thin_dir"]))
        stdout, stderr, _ = single.cmd_block()

        # Clean up our tar
        try:
            os.remove(trans_tar)
        except OSError:
            pass

        # Read in the JSON data and return the data structure
        try:
            return salt.utils.json.loads(stdout)
        except Exception as e:  # pylint: disable=broad-except
            log.error("JSON Render failed for: %s\n%s", stdout, stderr)
            log.error(str(e))

        # If for some reason the json load fails, return the stdout
        return stdout


def show_highstate(**kwargs):
    """
    Retrieve the highstate data from the salt master and display it

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_highstate
    """
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    pillar_override = kwargs.get("pillar")
    initial_pillar = _get_initial_pillar(opts)
    with salt.client.ssh.state.SSHHighState(
        opts,
        pillar_override,
        __salt__,
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:
        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += st_.opts["pillar"]["_errors"]
            return err
        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        st_.push_active()
        chunks = st_.compile_highstate(context=__context__.value())
        # Check for errors
        if not isinstance(chunks, dict):
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return chunks
        _cleanup_slsmod_high_data(chunks)
        return chunks


def show_lowstate(**kwargs):
    """
    List out the low data that will be applied to this minion

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_lowstate
    """
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    with salt.client.ssh.state.SSHHighState(
        opts,
        None,
        __salt__,
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=_get_initial_pillar(opts),
    ) as st_:
        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += st_.opts["pillar"]["_errors"]
            return err
        st_.push_active()
        chunks = st_.compile_low_chunks(context=__context__.value())
        _cleanup_slsmod_low_data(chunks)
        return chunks


def sls_id(id_, mods, test=None, queue=False, **kwargs):
    """
    Call a single ID from the named module(s) and handle all requisites

    The state ID comes *before* the module ID(s) on the command line.

    id
        ID to call

    mods
        Comma-delimited list of modules to search for given id and its requisites

    .. versionadded:: 2017.7.3

    saltenv : base
        Specify a salt fileserver environment to be used when applying states

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

    CLI Example:

    .. code-block:: bash

        salt '*' state.sls_id my_state my_module

        salt '*' state.sls_id my_state my_module,a_common_module
    """
    st_kwargs = __salt__.kwargs
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)

    # Since this is running a specific ID within a specific SLS file, fall back
    # to the 'base' saltenv if none is configured and none was passed.
    if opts["saltenv"] is None:
        opts["saltenv"] = "base"

    pillar_override = kwargs.get("pillar")
    initial_pillar = _get_initial_pillar(opts)
    with salt.client.ssh.state.SSHHighState(
        __opts__,
        pillar_override,
        __salt__,
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:

        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += __pillar__["_errors"]
            return err

        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        split_mods = _parse_mods(mods)
        st_.push_active()
        high_, errors = st_.render_highstate(
            {opts["saltenv"]: split_mods}, context=__context__.value()
        )
        errors += st_.state.verify_high(high_)
        # Apply requisites to high data
        high_, req_in_errors = st_.state.requisite_in(high_)
        if req_in_errors:
            # This if statement should not be necessary if there were no errors,
            # but it is required to get the unit tests to pass.
            errors.extend(req_in_errors)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        chunks = st_.state.compile_high_data(high_)
        chunk = [x for x in chunks if x.get("__id__", "") == id_]

        if not chunk:
            raise SaltInvocationError(
                "No matches for ID '{}' found in SLS '{}' within saltenv '{}'".format(
                    id_, mods, opts["saltenv"]
                )
            )

        ret = _ssh_state(chunk, st_kwargs, kwargs, pillar, test=test)
        _set_retcode(ret, highstate=highstate)
        # Work around Windows multiprocessing bug, set __opts__['test'] back to
        # value from before this function was run.
        __opts__["test"] = orig_test
        return ret


def show_sls(mods, saltenv="base", test=None, **kwargs):
    """
    Display the state data from a specific sls or list of sls files on the
    master

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_sls core,edit.vim dev
    """
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)
    pillar_override = kwargs.get("pillar")
    initial_pillar = _get_initial_pillar(opts)
    with salt.client.ssh.state.SSHHighState(
        opts,
        pillar_override,
        __salt__,
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:
        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += st_.opts["pillar"]["_errors"]
            return err
        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        st_.push_active()
        mods = _parse_mods(mods)
        high_data, errors = st_.render_highstate(
            {saltenv: mods}, context=__context__.value()
        )
        high_data, ext_errors = st_.state.reconcile_extend(high_data)
        errors += ext_errors
        errors += st_.state.verify_high(high_data)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        high_data, req_in_errors = st_.state.requisite_in(high_data)
        errors += req_in_errors
        high_data = st_.state.apply_exclude(high_data)
        # Verify that the high data is structurally sound
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        _cleanup_slsmod_high_data(high_data)
        return high_data


def show_low_sls(mods, saltenv="base", test=None, **kwargs):
    """
    Display the low state data from a specific sls or list of sls files on the
    master.

    .. versionadded:: 2016.3.2

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_low_sls core,edit.vim dev
    """
    __opts__["grains"] = __grains__.value()
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)
    pillar_override = kwargs.get("pillar")
    initial_pillar = _get_initial_pillar(opts)
    with salt.client.ssh.state.SSHHighState(
        opts,
        pillar_override,
        __salt__,
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=initial_pillar,
    ) as st_:
        if not _check_pillar(kwargs, st_.opts["pillar"]):
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            err = ["Pillar failed to render with the following messages:"]
            err += st_.opts["pillar"]["_errors"]
            return err
        try:
            pillar = st_.opts["pillar"].value()
        except AttributeError:
            pillar = st_.opts["pillar"]
        if pillar_override is not None or initial_pillar is None:
            # Ensure other wrappers use the correct pillar
            __pillar__.update(pillar)
        st_.push_active()
        mods = _parse_mods(mods)
        high_data, errors = st_.render_highstate(
            {saltenv: mods}, context=__context__.value()
        )
        high_data, ext_errors = st_.state.reconcile_extend(high_data)
        errors += ext_errors
        errors += st_.state.verify_high(high_data)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        high_data, req_in_errors = st_.state.requisite_in(high_data)
        errors += req_in_errors
        high_data = st_.state.apply_exclude(high_data)
        # Verify that the high data is structurally sound
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        ret = st_.state.compile_high_data(high_data)
        _cleanup_slsmod_low_data(ret)
        return ret


def sls_exists(mods, test=None, queue=None, **kwargs):
    """
    Tests for the existence of a specific SLS or list of SLS files on the
    master. Similar to :py:func:`state.show_sls <salt.modules.state.show_sls>`,
    rather than returning state details, returns True or False. The default
    environment is ``base``, use ``saltenv`` to specify a different environment.

    .. versionadded:: 3008.0

    saltenv
        Specify a salt fileserver environment from which to look for the SLS files
        specified in the ``mods`` argument

    CLI Example:

    .. code-block:: bash

        salt '*' state.sls_exists core,edit.vim saltenv=dev
    """
    return isinstance(show_sls(mods, test=test, queue=queue, **kwargs), dict)


def show_top(**kwargs):
    """
    Return the top data that the minion will use for a highstate

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_top
    """
    __opts__["grains"] = __grains__
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    with salt.client.ssh.state.SSHHighState(
        opts,
        None,
        __salt__,
        __context__["fileclient"],
        context=__context__.value(),
        initial_pillar=_get_initial_pillar(opts),
    ) as st_:
        top_data = st_.get_top(context=__context__.value())
        errors = []
        errors += st_.verify_tops(top_data)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        matches = st_.top_matches(top_data)
        return matches


def single(fun, name, test=None, **kwargs):
    """
    .. versionadded:: 2015.5.0

    Execute a single state function with the named kwargs, returns False if
    insufficient data is sent to the command

    By default, the values of the kwargs will be parsed as YAML. So, you can
    specify lists values, or lists of single entry key-value maps, as you
    would in a YAML salt file. Alternatively, JSON format of keyword values
    is also supported.

    CLI Example:

    .. code-block:: bash

        salt '*' state.single pkg.installed name=vim

    """
    st_kwargs = __salt__.kwargs
    __opts__["grains"] = __grains__.value()

    # state.fun -> [state, fun]
    comps = fun.split(".")
    if len(comps) < 2:
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return "Invalid function passed"

    # Create the low chunk, using kwargs as a base
    kwargs.update({"state": comps[0], "fun": comps[1], "__id__": name, "name": name})

    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)

    # Set test mode
    opts["test"] = _get_test_value(test, **kwargs)

    # Get the override pillar data
    # This needs to be removed from the kwargs, they are called
    # as a lowstate with one item, not a single chunk
    pillar_override = kwargs.pop("pillar", None)

    # Create the State environment
    st_ = salt.client.ssh.state.SSHState(
        opts, pillar_override, initial_pillar=_get_initial_pillar(opts)
    )

    try:
        pillar = st_.opts["pillar"].value()
    except AttributeError:
        pillar = st_.opts["pillar"]
    # Verify the low chunk
    err = st_.verify_data(kwargs)
    if err:
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return err

    # Must be a list of low-chunks
    chunks = [kwargs]

    # Retrieve file refs for the state run, so we can copy relevant files down
    # to the minion before executing the state
    file_refs = salt.client.ssh.state.lowstate_file_refs(
        chunks,
        _merge_extra_filerefs(
            kwargs.get("extra_filerefs", ""), opts.get("extra_filerefs", "")
        ),
    )

    roster = salt.roster.Roster(opts, opts.get("roster", "flat"))
    roster_grains = roster.opts["grains"]

    # Create the tar containing the state pkg and relevant files.
    trans_tar = salt.client.ssh.state.prep_trans_tar(
        __context__["fileclient"],
        chunks,
        file_refs,
        pillar,
        st_kwargs["id_"],
        roster_grains,
    )

    # Create a hash so we can verify the tar on the target system
    trans_tar_sum = salt.utils.hashutils.get_hash(trans_tar, opts["hash_type"])

    # We use state.pkg to execute the "state package"
    cmd = "state.pkg {}/salt_state.tgz test={} pkg_sum={} hash_type={}".format(
        opts["thin_dir"], test, trans_tar_sum, opts["hash_type"]
    )

    # Create a salt-ssh Single object to actually do the ssh work
    single = salt.client.ssh.Single(
        opts,
        cmd,
        fsclient=__context__["fileclient"],
        minion_opts=__salt__.minion_opts,
        **st_kwargs,
    )

    # Copy the tar down
    single.shell.send(trans_tar, "{}/salt_state.tgz".format(opts["thin_dir"]))

    # Run the state.pkg command on the target
    stdout, stderr, _ = single.cmd_block()

    # Clean up our tar
    try:
        os.remove(trans_tar)
    except OSError:
        pass

    # Read in the JSON data and return the data structure
    try:
        return salt.utils.json.loads(stdout)
    except Exception as e:  # pylint: disable=broad-except
        log.error("JSON Render failed for: %s\n%s", stdout, stderr)
        log.error(str(e))

    # If for some reason the json load fails, return the stdout
    return stdout


def test(*args, **kwargs):
    """
    .. versionadded:: 3001

    Alias for `state.apply` with the kwarg `test` forced to `True`.

    This is a nicety to avoid the need to type out `test=True` and the possibility of
    a typo causing changes you do not intend.
    """
    kwargs["test"] = True
    ret = apply_(*args, **kwargs)

    return ret



"""
tests.pytests.unit.test_crypt
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for salt's crypt module
"""

import uuid

import pytest

import salt.crypt
import salt.master
import salt.utils.files
from tests.conftest import FIPS_TESTRUN

PRIV_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAoAsMPt+4kuIG6vKyw9r3+OuZrVBee/2vDdVetW+Js5dTlgrJ
aghWWn3doGmKlEjqh7E4UTa+t2Jd6w8RSLnyHNJ/HpVhMG0M07MF6FMfILtDrrt8
ZX7eDVt8sx5gCEpYI+XG8Y07Ga9i3Hiczt+fu6HYwu96HggmG2pqkOrn3iGfqBvV
YVFJzSZYe7e4c1PeEs0xYcrA4k+apyGsMtpef8vRUrNicRLc7dAcvfhtgt2DXEZ2
d72t/CR4ygtUvPXzisaTPW0G7OWAheCloqvTIIPQIjR8htFxGTz02STVXfnhnJ0Z
k8KhqKF2v1SQvIYxsZU7jaDgl5i3zpeh58cYOwIDAQABAoIBABZUJEO7Y91+UnfC
H6XKrZEZkcnH7j6/UIaOD9YhdyVKxhsnax1zh1S9vceNIgv5NltzIsfV6vrb6v2K
Dx/F7Z0O0zR5o+MlO8ZncjoNKskex10gBEWG00Uqz/WPlddiQ/TSMJTv3uCBAzp+
S2Zjdb4wYPUlgzSgb2ygxrhsRahMcSMG9PoX6klxMXFKMD1JxiY8QfAHahPzQXy9
F7COZ0fCVo6BE+MqNuQ8tZeIxu8mOULQCCkLFwXmkz1FpfK/kNRmhIyhxwvCS+z4
JuErW3uXfE64RLERiLp1bSxlDdpvRO2R41HAoNELTsKXJOEt4JANRHm/CeyA5wsh
NpscufUCgYEAxhgPfcMDy2v3nL6KtkgYjdcOyRvsAF50QRbEa8ldO+87IoMDD/Oe
osFERJ5hhyyEO78QnaLVegnykiw5DWEF02RKMhD/4XU+1UYVhY0wJjKQIBadsufB
2dnaKjvwzUhPh5BrBqNHl/FXwNCRDiYqXa79eWCPC9OFbZcUWWq70s8CgYEAztOI
61zRfmXJ7f70GgYbHg+GA7IrsAcsGRITsFR82Ho0lqdFFCxz7oK8QfL6bwMCGKyk
nzk+twh6hhj5UNp18KN8wktlo02zTgzgemHwaLa2cd6xKgmAyuPiTgcgnzt5LVNG
FOjIWkLwSlpkDTl7ZzY2QSy7t+mq5d750fpIrtUCgYBWXZUbcpPL88WgDB7z/Bjg
dlvW6JqLSqMK4b8/cyp4AARbNp12LfQC55o5BIhm48y/M70tzRmfvIiKnEc/gwaE
NJx4mZrGFFURrR2i/Xx5mt/lbZbRsmN89JM+iKWjCpzJ8PgIi9Wh9DIbOZOUhKVB
9RJEAgo70LvCnPTdS0CaVwKBgDJW3BllAvw/rBFIH4OB/vGnF5gosmdqp3oGo1Ik
jipmPAx6895AH4tquIVYrUl9svHsezjhxvjnkGK5C115foEuWXw0u60uiTiy+6Pt
2IS0C93VNMulenpnUrppE7CN2iWFAiaura0CY9fE/lsVpYpucHAWgi32Kok+ZxGL
WEttAoGAN9Ehsz4LeQxEj3x8wVeEMHF6OsznpwYsI2oVh6VxpS4AjgKYqeLVcnNi
TlZFsuQcqgod8OgzA91tdB+Rp86NygmWD5WzeKXpCOg9uA+y/YL+0sgZZHsuvbK6
PllUgXdYxqClk/hdBFB7v9AQoaj7K9Ga22v32msftYDQRJ94xOI=
-----END RSA PRIVATE KEY-----
"""


PUB_KEY = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoAsMPt+4kuIG6vKyw9r3
+OuZrVBee/2vDdVetW+Js5dTlgrJaghWWn3doGmKlEjqh7E4UTa+t2Jd6w8RSLny
HNJ/HpVhMG0M07MF6FMfILtDrrt8ZX7eDVt8sx5gCEpYI+XG8Y07Ga9i3Hiczt+f
u6HYwu96HggmG2pqkOrn3iGfqBvVYVFJzSZYe7e4c1PeEs0xYcrA4k+apyGsMtpe
f8vRUrNicRLc7dAcvfhtgt2DXEZ2d72t/CR4ygtUvPXzisaTPW0G7OWAheCloqvT
IIPQIjR8htFxGTz02STVXfnhnJ0Zk8KhqKF2v1SQvIYxsZU7jaDgl5i3zpeh58cY
OwIDAQAB
-----END PUBLIC KEY-----
"""

PRIV_KEY2 = """
-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAp+8cTxguO6Vg+YO92VfHgNld3Zy8aM3JbZvpJcjTnis+YFJ7
Zlkcc647yPRRwY9nYBNywahnt5kIeuT1rTvTsMBZWvmUoEVUj1Xg8XXQkBvb9Ozy
Gqy/G/p8KDDpzMP/U+XCnUeHiXTZrgnqgBIc2cKeCVvWFqDi0GRFGzyaXLaX3PPm
M7DJ0MIPL1qgmcDq6+7Ze0gJ9SrDYFAeLmbuT1OqDfufXWQl/82JXeiwU2cOpqWq
7n5fvPOWim7l1tzQ+dSiMRRm0xa6uNexCJww3oJSwvMbAmgzvOhqqhlqv+K7u0u7
FrFFojESsL36Gq4GBrISnvu2tk7u4GGNTYYQbQIDAQABAoIBAADrqWDQnd5DVZEA
lR+WINiWuHJAy/KaIC7K4kAMBgbxrz2ZbiY9Ok/zBk5fcnxIZDVtXd1sZicmPlro
GuWodIxdPZAnWpZ3UtOXUayZK/vCP1YsH1agmEqXuKsCu6Fc+K8VzReOHxLUkmXn
FYM+tixGahXcjEOi/aNNTWitEB6OemRM1UeLJFzRcfyXiqzHpHCIZwBpTUAsmzcG
QiVDkMTKubwo/m+PVXburX2CGibUydctgbrYIc7EJvyx/cpRiPZXo1PhHQWdu4Y1
SOaC66WLsP/wqvtHo58JQ6EN/gjSsbAgGGVkZ1xMo66nR+pLpR27coS7o03xCks6
DY/0mukCgYEAuLIGgBnqoh7YsOBLd/Bc1UTfDMxJhNseo+hZemtkSXz2Jn51322F
Zw/FVN4ArXgluH+XsOhvG/MFFpojwZSrb0Qq5b1MRdo9qycq8lGqNtlN1WHqosDQ
zW29kpL0tlRrSDpww3wRESsN9rH5XIrJ1b3ZXuO7asR+KBVQMy/+NcUCgYEA6MSC
c+fywltKPgmPl5j0DPoDe5SXE/6JQy7w/vVGrGfWGf/zEJmhzS2R+CcfTTEqaT0T
Yw8+XbFgKAqsxwtE9MUXLTVLI3sSUyE4g7blCYscOqhZ8ItCUKDXWkSpt++rG0Um
1+cEJP/0oCazG6MWqvBC4NpQ1nzh46QpjWqMwokCgYAKDLXJ1p8rvx3vUeUJW6zR
dfPlEGCXuAyMwqHLxXgpf4EtSwhC5gSyPOtx2LqUtcrnpRmt6JfTH4ARYMW9TMef
QEhNQ+WYj213mKP/l235mg1gJPnNbUxvQR9lkFV8bk+AGJ32JRQQqRUTbU+yN2MQ
HEptnVqfTp3GtJIultfwOQKBgG+RyYmu8wBP650izg33BXu21raEeYne5oIqXN+I
R5DZ0JjzwtkBGroTDrVoYyuH1nFNEh7YLqeQHqvyufBKKYo9cid8NQDTu+vWr5UK
tGvHnwdKrJmM1oN5JOAiq0r7+QMAOWchVy449VNSWWV03aeftB685iR5BXkstbIQ
EVopAoGAfcGBTAhmceK/4Q83H/FXBWy0PAa1kZGg/q8+Z0KY76AqyxOVl0/CU/rB
3tO3sKhaMTHPME/MiQjQQGoaK1JgPY6JHYvly2KomrJ8QTugqNGyMzdVJkXAK2AM
GAwC8ivAkHf8CHrHa1W7l8t2IqBjW1aRt7mOW92nfG88Hck0Mbo=
-----END RSA PRIVATE KEY-----
"""


PUB_KEY2 = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAp+8cTxguO6Vg+YO92VfH
gNld3Zy8aM3JbZvpJcjTnis+YFJ7Zlkcc647yPRRwY9nYBNywahnt5kIeuT1rTvT
sMBZWvmUoEVUj1Xg8XXQkBvb9OzyGqy/G/p8KDDpzMP/U+XCnUeHiXTZrgnqgBIc
2cKeCVvWFqDi0GRFGzyaXLaX3PPmM7DJ0MIPL1qgmcDq6+7Ze0gJ9SrDYFAeLmbu
T1OqDfufXWQl/82JXeiwU2cOpqWq7n5fvPOWim7l1tzQ+dSiMRRm0xa6uNexCJww
3oJSwvMbAmgzvOhqqhlqv+K7u0u7FrFFojESsL36Gq4GBrISnvu2tk7u4GGNTYYQ
bQIDAQAB
-----END PUBLIC KEY-----
"""


def test_get_rsa_pub_key_bad_key(tmp_path):
    """
    get_rsa_pub_key raises InvalidKeyError when encoutering a bad key
    """
    key_path = str(tmp_path / "key")
    with salt.utils.files.fopen(key_path, "w") as fp:
        fp.write("")
    with pytest.raises(salt.crypt.InvalidKeyError):
        salt.crypt.get_rsa_pub_key(key_path)


def test_cryptical_dumps_no_nonce():
    master_crypt = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())
    data = {"foo": "bar"}
    ret = master_crypt.dumps(data)

    # Validate message structure
    assert isinstance(ret, bytes)
    une = master_crypt.decrypt(ret)
    une.startswith(master_crypt.PICKLE_PAD)
    assert salt.payload.loads(une[len(master_crypt.PICKLE_PAD) :]) == data

    # Validate load back to orig data
    assert master_crypt.loads(ret) == data


def test_cryptical_dumps_valid_nonce():
    nonce = uuid.uuid4().hex
    master_crypt = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())
    data = {"foo": "bar"}
    ret = master_crypt.dumps(data, nonce=nonce)

    assert isinstance(ret, bytes)
    une = master_crypt.decrypt(ret)
    une.startswith(master_crypt.PICKLE_PAD)
    nonce_and_data = une[len(master_crypt.PICKLE_PAD) :]
    assert nonce_and_data.startswith(nonce.encode())
    assert salt.payload.loads(nonce_and_data[len(nonce) :]) == data

    assert master_crypt.loads(ret, nonce=nonce) == data


def test_cryptical_dumps_invalid_nonce():
    nonce = uuid.uuid4().hex
    master_crypt = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())
    data = {"foo": "bar"}
    ret = master_crypt.dumps(data, nonce=nonce)
    assert isinstance(ret, bytes)
    with pytest.raises(salt.crypt.SaltClientError, match="Nonce verification error"):
        assert master_crypt.loads(ret, nonce="abcde")


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_verify_signature(tmp_path):
    tmp_path.joinpath("foo.pem").write_text(PRIV_KEY.strip())
    tmp_path.joinpath("foo.pub").write_text(PUB_KEY.strip())
    tmp_path.joinpath("bar.pem").write_text(PRIV_KEY2.strip())
    tmp_path.joinpath("bar.pub").write_text(PUB_KEY2.strip())
    msg = b"foo bar"
    sig = salt.crypt.sign_message(str(tmp_path.joinpath("foo.pem")), msg)
    assert salt.crypt.verify_signature(str(tmp_path.joinpath("foo.pub")), msg, sig)


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_verify_signature_bad_sig(tmp_path):
    tmp_path.joinpath("foo.pem").write_text(PRIV_KEY.strip())
    tmp_path.joinpath("foo.pub").write_text(PUB_KEY.strip())
    tmp_path.joinpath("bar.pem").write_text(PRIV_KEY2.strip())
    tmp_path.joinpath("bar.pub").write_text(PUB_KEY2.strip())
    msg = b"foo bar"
    sig = salt.crypt.sign_message(str(tmp_path.joinpath("foo.pem")), msg)
    assert not salt.crypt.verify_signature(str(tmp_path.joinpath("bar.pub")), msg, sig)



# Python libs

import logging

import pytest

# Salt libs
import salt.beacons.cert_info as cert_info
from tests.support.mock import mock_open, patch

log = logging.getLogger(__name__)


_TEST_CERT = """
-----BEGIN CERTIFICATE-----
MIIC/jCCAeagAwIBAgIJAIQMfu6ShHvfMA0GCSqGSIb3DQEBCwUAMCQxIjAgBgNV
BAMMGXNhbHR0ZXN0LTAxLmV4YW1wbGUubG9jYWwwHhcNMTkwNjAzMjA1OTIyWhcN
MjkwNTMxMjA1OTIyWjAkMSIwIAYDVQQDDBlzYWx0dGVzdC0wMS5leGFtcGxlLmxv
Y2FsMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAv5UxxKGsOO8n2hUk
KjL8r2Rjt0om4wwdXUu0R1fQUlaSO0g+vk0wHHaovoVcEU6uZlhDPw1qZ4C+cp9Z
rDzSfwI2Njg813I5jzTBgox+3pJ+82vgXZ14xpqZ+f0ACMo4uRPjBkyQpHqYiDJ3
VockZSxm5s7RT05xDnedDfPgu1WAvzQovWO6slCs+Hlp8sh6QAy/hIwOZ0hT8y3J
NV6PSPqK7BEypOPak36+ogtiuPxxat4da74SUVS8Ffupnr40BjqVqEXBvfIIHiQt
3r5gpjoBjrWX2ccgQlHQP8gFaToFxWLSSYVT6E8Oj5UEywpmvPDRjJsJ5epscblT
oFyVXQIDAQABozMwMTAJBgNVHRMEAjAAMCQGA1UdEQQdMBuCGXNhbHR0ZXN0LTAx
LmV4YW1wbGUubG9jYWwwDQYJKoZIhvcNAQELBQADggEBABPqQlkaZDV5dPwNO/s2
PBT/19LroOwQ+fBJgZpbGha5/ZaSr+jcYZf2jAicPajWGlY/rXAdBSuxpmUYCC12
23tI4stwGyB8Quuoyg2Z+5LQJSDA1LxNJ1kxQfDUnS3tVQa0wJVtq8W9wNryNONL
noaQaDcdbGx3V15W+Bx0as5NfIWqz1uVi4MGGxI6hMBuDD7E7M+k1db8EaS+tI4u
seZBENjwjJA6zZmTXvYyzV5OBP4JyOhYuG9aqr7e6/yjPBEtZv0TJ9KMMbcywvE9
9FF+l4Y+wgKR/icrpDEpPlC4wYn64sy5vk7EGVagnVyhkjLJ52rn4trzyPox8FmO
2Zw=
-----END CERTIFICATE-----
"""


@pytest.fixture
def configure_loader_modules():
    return {cert_info: {"__context__": {}, "__salt__": {}}}


def test_non_list_config():
    config = {}

    ret = cert_info.validate(config)

    assert ret == (False, "Configuration for cert_info beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = cert_info.validate(config)

    assert ret == (
        False,
        "Configuration for cert_info beacon must contain files option.",
    )


def test_cert_information():
    with patch("salt.utils.files.fopen", mock_open(read_data=_TEST_CERT)):
        config = [{"files": ["/etc/pki/tls/certs/mycert.pem"], "notify_days": -1}]

        ret = cert_info.validate(config)

        assert ret == (True, "Valid beacon configuration")

        _expected_return = [
            {
                "certificates": [
                    {
                        "cert_path": "/etc/pki/tls/certs/mycert.pem",
                        "extensions": [
                            {
                                "ext_data": "CA:FALSE",
                                "ext_name": "basicConstraints",
                            },
                            {
                                "ext_data": "DNS:salttest-01.example.local",
                                "ext_name": "subjectAltName",
                            },
                        ],
                        "has_expired": False,
                        "issuer": 'CN="salttest-01.example.local"',
                        "issuer_dict": {"CN": "salttest-01.example.local"},
                        "notAfter": "2029-05-31 20:59:22Z",
                        "notAfter_raw": "20290531205922Z",
                        "notBefore": "2019-06-03 20:59:22Z",
                        "notBefore_raw": "20190603205922Z",
                        "serial_number": 9515119675852487647,
                        "signature_algorithm": "sha256WithRSAEncryption",
                        "subject": 'CN="salttest-01.example.local"',
                        "subject_dict": {"CN": "salttest-01.example.local"},
                        "version": 2,
                    }
                ]
            }
        ]
        ret = cert_info.beacon(config)
        assert ret == _expected_return



import ctypes
import multiprocessing
import uuid

import pytest

import salt.channel.server as server
import salt.crypt
import salt.daemons.masterapi
import salt.master
import salt.payload
import salt.utils.event
import salt.utils.files
import salt.utils.stringutils
from salt.master import SMaster


@pytest.fixture
def key_data():
    return [
        "-----BEGIN PUBLIC KEY-----",
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoe5QSDYRWKyknbVyRrIj",
        "rm1ht5HgKzAVUber0x54+b/UgxTd1cqI6I+eDlx53LqZSH3G8Rd5cUh8LHoGedSa",
        "E62vEiLAjgXa+RdgcGiQpYS8+Z2RvQJ8oIcZgO+2AzgBRHboNWHTYRRmJXCd3dKs",
        "9tcwK6wxChR06HzGqaOTixAuQlegWbOTU+X4dXIbW7AnuQBt9MCib7SxHlscrqcS",
        "cBrRvq51YP6cxPm/rZJdBqZhVrlghBvIpa45NApP5PherGi4AbEGYte4l+gC+fOA",
        "osEBis1V27djPpIyQS4qk3XAPQg6CYQMDltHqA4Fdo0Nt7SMScxJhfH0r6zmBFAe",
        "BQIDAQAB",
        "-----END PUBLIC KEY-----",
    ]


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test_compare_keys(key_data, linesep):
    src_key = linesep.join(key_data)
    tgt_key = "\n".join(key_data)
    assert server.ReqServerChannel.compare_keys(src_key, tgt_key) is True


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test_compare_keys_newline_src(key_data, linesep):
    src_key = linesep.join(key_data) + linesep
    tgt_key = "\n".join(key_data)
    assert src_key.endswith(linesep)
    assert not tgt_key.endswith("\n")
    assert server.ReqServerChannel.compare_keys(src_key, tgt_key) is True


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test_compare_keys_newline_tgt(key_data, linesep):
    src_key = linesep.join(key_data)
    tgt_key = "\n".join(key_data) + "\n"
    assert not src_key.endswith(linesep)
    assert tgt_key.endswith("\n")
    assert server.ReqServerChannel.compare_keys(src_key, tgt_key) is True


@pytest.fixture
def root_dir(tmp_path):
    (tmp_path / "var").mkdir()
    (tmp_path / "var" / "cache").mkdir()
    (tmp_path / "etc").mkdir()
    (tmp_path / "etc" / "salt").mkdir()
    (tmp_path / "etc" / "salt" / "pki").mkdir()
    (tmp_path / "etc" / "salt" / "pki" / "minions").mkdir()
    yield tmp_path


def test_req_server_validate_token_removes_token(root_dir):
    opts = {
        "master_uri": "tcp://127.0.0.1:4505",
        "cachedir": str(root_dir / "var" / "cache"),
        "pki_dir": str(root_dir / "etc" / "salt" / "pki"),
    }
    reqsrv = server.ReqServerChannel.factory(opts)
    payload = {
        "load": {
            "id": "minion",
            "tok": "asdf",
        }
    }
    assert reqsrv.validate_token(payload) is False
    assert "tok" not in payload["load"]


def test_req_server_validate_token_removes_token_id_traversal(root_dir):
    opts = {
        "master_uri": "tcp://127.0.0.1:4505",
        "cachedir": str(root_dir / "var" / "cache"),
        "pki_dir": str(root_dir / "etc" / "salt" / "pki"),
    }
    reqsrv = server.ReqServerChannel.factory(opts)
    payload = {
        "load": {
            "id": "minion/../../foo",
            "tok": "asdf",
        }
    }
    assert reqsrv.validate_token(payload) is False
    assert "tok" not in payload["load"]


# ============================================================================
# Auth Version Downgrade Attack Regression Tests
# ============================================================================


@pytest.fixture
def pki_dir(tmp_path):
    """Setup PKI directory structure for auth tests."""
    master_pki = tmp_path / "master"
    minion_pki = tmp_path / "minion"

    master_pki.mkdir()
    minion_pki.mkdir()
    (master_pki / "minions").mkdir()
    (master_pki / "minions_pre").mkdir()
    (master_pki / "minions_rejected").mkdir()
    (master_pki / "minions_denied").mkdir()

    # Generate master keys
    salt.crypt.gen_keys(str(master_pki), "master", 4096)

    # Generate minion keys
    salt.crypt.gen_keys(str(minion_pki), "minion", 4096)

    return tmp_path


@pytest.fixture
def auth_master_opts(pki_dir, tmp_path):
    """Master configuration for auth tests."""
    opts = {
        "master_uri": "tcp://127.0.0.1:4506",
        "interface": "127.0.0.1",
        "ret_port": 4506,
        "ipv6": False,
        "sock_dir": str(tmp_path / "sock"),
        "cachedir": str(tmp_path / "cache"),
        "pki_dir": str(pki_dir / "master"),
        "id": "master",
        "__role": "master",
        "keysize": 4096,
        "max_minions": 0,
        "auto_accept": False,
        "open_mode": False,
        "key_pass": None,
        "publish_port": 4505,
        "auth_mode": 1,
        "auth_events": True,
        "publish_session": 86400,
        "request_server_ttl": 300,  # 5 minutes
        "master_sign_pubkey": False,
        "sign_pub_messages": False,
        "cluster_id": None,
        "transport": "zeromq",
        "minimum_auth_version": 3,  # Enforce version 3+ for security
    }
    (tmp_path / "sock").mkdir(exist_ok=True)
    (tmp_path / "cache").mkdir(exist_ok=True)
    (tmp_path / "cache" / "sessions").mkdir(exist_ok=True)
    return opts


@pytest.fixture
def auth_minion_opts(pki_dir, auth_master_opts):
    """Minion configuration for auth tests."""
    return {
        "master_uri": "tcp://127.0.0.1:4506",
        "interface": "127.0.0.1",
        "ret_port": 4506,
        "ipv6": False,
        "sock_dir": str(auth_master_opts["sock_dir"]),
        "pki_dir": str(pki_dir / "minion"),
        "id": "test-minion",
        "__role": "minion",
        "keysize": 4096,
        "encryption_algorithm": salt.crypt.OAEP_SHA1,
        "signing_algorithm": salt.crypt.PKCS1v15_SHA1,
    }


@pytest.fixture
def setup_accepted_minion(pki_dir, auth_minion_opts, auth_master_opts):
    """
    Setup a pre-accepted minion by copying its public key to master's
    minions directory.
    """
    minion_pub = pki_dir / "minion" / "minion.pub"
    master_minions_dir = pki_dir / "master" / "minions"
    accepted_key = master_minions_dir / auth_minion_opts["id"]

    # Copy minion public key to master's accepted keys
    with salt.utils.files.fopen(str(minion_pub)) as src:
        with salt.utils.files.fopen(str(accepted_key), "w") as dst:
            dst.write(src.read())

    return accepted_key


@pytest.fixture
def req_server(auth_master_opts):
    """Create a ReqServerChannel instance for testing."""
    # Setup master secrets
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }

    server_channel = server.ReqServerChannel.factory(auth_master_opts)
    server_channel.auto_key = salt.daemons.masterapi.AutoKey(auth_master_opts)
    server_channel.cache_cli = False
    server_channel.event = salt.utils.event.get_master_event(
        auth_master_opts, auth_master_opts["sock_dir"], listen=False
    )
    server_channel.master_key = salt.crypt.MasterKeys(auth_master_opts)

    yield server_channel

    server_channel.close()
    if "aes" in SMaster.secrets:
        SMaster.secrets.pop("aes")


async def test_auth_version_downgrade_v3_to_v0(
    pki_dir, auth_minion_opts, req_server, setup_accepted_minion
):
    """
    REGRESSION TEST: CVE-2025-62349 - Auth Version Downgrade Attack

    Test that the master rejects authentication attempts using protocol version 0
    when minimum_auth_version is set to 3.

    This prevents a malicious or compromised minion from using an older protocol
    version to bypass security features introduced in version 3+ (token validation,
    TTL checks, ID matching, session keys).

    The test verifies that the minimum_auth_version enforcement works at the initial
    authentication stage, preventing minions from establishing low-version sessions.
    """
    # Read minion public key
    with salt.utils.files.fopen(str(pki_dir / "minion" / "minion.pub"), "r") as fp:
        pub_key = fp.read()

    # Simulate version 0 auth load (no nonce, no signing)
    load_v0 = {
        "cmd": "_auth",
        "id": auth_minion_opts["id"],
        "pub": pub_key,
        "enc_algo": auth_minion_opts["encryption_algorithm"],
        "sig_algo": auth_minion_opts["signing_algorithm"],
    }

    # Create payload for handle_message (version 0)
    payload = {"enc": "clear", "load": load_v0, "version": 0}

    # Call handle_message which will enforce minimum_auth_version
    ret = await req_server.handle_message(payload)

    # REGRESSION TEST: Version 0 should be REJECTED
    # With minimum_auth_version=3, version 0 should return "bad load"
    assert ret == "bad load", "Expected 'bad load' for rejected version 0 auth"


@pytest.mark.parametrize("downgrade_version", [0, 1, 2])
async def test_auth_version_downgrade_from_v3(
    pki_dir, auth_minion_opts, req_server, setup_accepted_minion, downgrade_version
):
    """
    REGRESSION TEST: CVE-2025-62349 - Auth Version Downgrade Attack (Parametrized)

    Test that the master rejects authentication attempts using protocol versions
    0, 1, or 2 when minimum_auth_version is set to 3.

    This prevents malicious or compromised minions from using older protocol
    versions to bypass security features:
    - v0/v1: No message signing, no nonce
    - v2: Message signing but no token validation, no TTL checks, no ID matching
    - v3+: Full security (token validation, TTL, ID matching, session keys)

    The test verifies that minimum_auth_version enforcement prevents minions from
    establishing low-version sessions that would allow them to bypass security
    controls and potentially impersonate other minions or maintain unauthorized access.
    """
    with salt.utils.files.fopen(str(pki_dir / "minion" / "minion.pub"), "r") as fp:
        pub_key = fp.read()

    load = {
        "cmd": "_auth",
        "id": auth_minion_opts["id"],
        "pub": pub_key,
        "enc_algo": auth_minion_opts["encryption_algorithm"],
        "sig_algo": auth_minion_opts["signing_algorithm"],
    }

    # Add nonce for v2+ (but not for v0/v1)
    if downgrade_version >= 2:
        load["nonce"] = uuid.uuid4().hex

    # Create payload for handle_message
    payload = {"enc": "clear", "load": load, "version": downgrade_version}

    # Call handle_message which will enforce minimum_auth_version
    ret = await req_server.handle_message(payload)

    # REGRESSION TEST: Old versions should be REJECTED
    # With minimum_auth_version=3, versions 0, 1, 2 should return "bad load"
    assert (
        ret == "bad load"
    ), f"Expected 'bad load' for rejected version {downgrade_version} auth"


def test_handle_message_version_extraction(auth_master_opts):
    """
    REGRESSION TEST: CVE-2025-62349 - Version Extraction from Untrusted Payload

    Test that the master should have minimum_auth_version configured.

    EXPECTED BEHAVIOR:
    - Master opts should have minimum_auth_version set
    - The commented code at salt/channel/server.py:144-145 should be enabled

    CURRENT BEHAVIOR (VULNERABLE):
    - Test will FAIL - minimum_auth_version is not in opts
    - After fix is implemented, this test will PASS
    """
    # The current code at salt/channel/server.py:139-145 shows:
    # version = payload.get("version", 0)
    # #if version < self.opts["minimum_auth_version"]:
    # #    raise salt.ext.tornado.gen.Return("bad load")

    # REGRESSION TEST: Verify minimum_auth_version exists in opts
    # Currently this will FAIL because the option doesn't exist
    assert (
        "minimum_auth_version" in auth_master_opts
    ), "Expected minimum_auth_version to be configured in master opts"
    assert (
        auth_master_opts["minimum_auth_version"] >= 3
    ), "Expected minimum auth version to be at least 3"


# Note: The remaining security bypasses (token, TTL, ID mismatch, session keys)
# are already tested via the parametrized downgrade tests above and the
# functional tests. The key regression test is ensuring old versions are rejected.



"""
    :codeauthor: Thomas Stoner <tmstoner@cisco.com>
"""

# Copyright (c) 2018 Cisco and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from tests.unit.modules.nxos.nxos_platform import NXOSPlatform


class N36KPlatform(NXOSPlatform):
    """Cisco Systems N36K Platform Unit Test Object"""

    chassis = "Nexus3000 N3K-C36180YC-R Chassis"

    # Captured output from: show install all impact nxos <image>

    show_install_all_impact = """
Installer will perform impact only check. Please wait.

Verifying image bootflash:/$IMAGE for boot variable "nxos".
[####################] 100% -- SUCCESS

Verifying image type.
[####################] 100% -- SUCCESS

Preparing "nxos" version info using image bootflash:/$IMAGE.
[####################] 100% -- SUCCESS

Preparing "bios" version info using image bootflash:/$IMAGE.
[####################] 100% -- SUCCESS

Performing module support checks.
[####################] 100% -- SUCCESS

Notifying services about system upgrade.
[####################] 100% -- SUCCESS



Compatibility check is done:
Module  bootable          Impact  Install-type  Reason
------  --------  --------------  ------------  ------
     1       yes      disruptive         reset  default upgrade is not hitless



Images will be upgraded according to following table:
Module       Image                  Running-Version(pri:alt)           New-Version  Upg-Required
------  ----------  ----------------------------------------  --------------------  ------------
     1        nxos                                    $CVER         $NVER           $REQ
     1        bios     v01.13(09/27/2018):v01.08(08/22/2017)    v01.13(09/27/2018)            no
    """

    # Captured output from: install all nxos <image>

    install_all_disruptive_success = """
Installer will perform compatibility check first. Please wait.
Installer is forced disruptive

Verifying image bootflash:/$IMAGE for boot variable "nxos".
[####################] 100% -- SUCCESS

Verifying image type.
[####################] 100% -- SUCCESS

Preparing "nxos" version info using image bootflash:/$IMAGE.
[####################] 100% -- SUCCESS

Preparing "bios" version info using image bootflash:/$IMAGE.
[####################] 100% -- SUCCESS

Performing module support checks.
[####################] 100% -- SUCCESS

Notifying services about system upgrade.
[####################] 100% -- SUCCESS



Compatibility check is done:
Module  bootable          Impact  Install-type  Reason
------  --------  --------------  ------------  ------
     1       yes      disruptive         reset  default upgrade is not hitless



Images will be upgraded according to following table:
Module       Image                  Running-Version(pri:alt)           New-Version  Upg-Required
------  ----------  ----------------------------------------  --------------------  ------------
     1        nxos                                    $CVER         $NVER           $REQ
     1        bios     v01.13(09/27/2018):v01.08(08/22/2017)    v01.13(09/27/2018)            no


Install is in progress, please wait.

Performing runtime checks.
[####################] 100% -- SUCCESS

Setting boot variables.
[####################] 100% -- SUCCESS

Performing configuration copy.
[####################] 100% -- SUCCESS

Module 1: Refreshing compact flash and upgrading bios/loader/bootrom.
Warning: please do not remove or power off the module at this time.
[####################] 100% -- SUCCESS


Finishing the upgrade, switch will reboot in 10 seconds.
    """



"""
salt-ssh testing
"""

import pathlib
import shutil

import pytest

import salt.utils.files
import salt.utils.yaml
from salt.defaults.exitcodes import EX_AGGREGATE

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(autouse=True)
def thin_dir(salt_ssh_cli):
    try:
        yield
    finally:
        ret = salt_ssh_cli.run("config.get", "thin_dir")
        assert ret.returncode == 0
        thin_dir_path = ret.data
        shutil.rmtree(thin_dir_path, ignore_errors=True)


def test_ping(salt_ssh_cli):
    """
    Test a simple ping
    """
    ret = salt_ssh_cli.run("test.ping")
    assert ret.returncode == 0
    assert ret.data is True


def test_thin_dir(salt_ssh_cli):
    """
    test to make sure thin_dir is created
    and salt-call file is included
    """
    ret = salt_ssh_cli.run("config.get", "thin_dir")
    assert ret.returncode == 0
    thin_dir = pathlib.Path(ret.data)
    assert thin_dir.is_dir()
    assert thin_dir.joinpath("salt-call").exists()
    assert thin_dir.joinpath("running_data").exists()


def test_wipe(salt_ssh_cli):
    """
    Ensure --wipe is respected by the state module wrapper
    issue #61083
    """
    ret = salt_ssh_cli.run("config.get", "thin_dir")
    assert ret.returncode == 0
    thin_dir = pathlib.Path(ret.data)
    assert thin_dir.exists()
    # only few modules (state and cp) will actually respect --wipe
    # (see commit #8a414d53284ec04940540ebd823306ab5119e105)
    salt_ssh_cli.run("--wipe", "state.apply")
    assert not thin_dir.exists()


def test_set_path(salt_ssh_cli, tmp_path, salt_ssh_roster_file):
    """
    test setting the path env variable
    """
    path = "/pathdoesnotexist/"
    roster_file = tmp_path / "roster-set-path"
    with salt.utils.files.fopen(salt_ssh_roster_file) as rfh:
        roster_data = salt.utils.yaml.safe_load(rfh)
        roster_data["localhost"].update(
            {
                "set_path": f"$PATH:/usr/local/bin/:{path}",
            }
        )
    with salt.utils.files.fopen(roster_file, "w") as wfh:
        salt.utils.yaml.safe_dump(roster_data, wfh)

    ret = salt_ssh_cli.run(f"--roster-file={roster_file}", "environ.get", "PATH")
    assert ret.returncode == 0
    assert path in ret.data


def test_tty(salt_ssh_cli, tmp_path, salt_ssh_roster_file):
    """
    test using tty
    """
    roster_file = tmp_path / "roster-tty"
    with salt.utils.files.fopen(salt_ssh_roster_file) as rfh:
        roster_data = salt.utils.yaml.safe_load(rfh)
        roster_data["localhost"].update({"tty": True})
    with salt.utils.files.fopen(roster_file, "w") as wfh:
        salt.utils.yaml.safe_dump(roster_data, wfh)
    ret = salt_ssh_cli.run(f"--roster-file={roster_file}", "test.ping")
    assert ret.returncode == 0
    assert ret.data is True


def test_retcode_exe_run_fail(salt_ssh_cli):
    """
    Verify salt-ssh passes through the retcode it receives.
    """
    ret = salt_ssh_cli.run("file.touch", "/tmp/non/ex/is/tent")
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, dict)
    assert "Error running 'file.touch': No such file or directory" in ret.data["stderr"]
    assert ret.data["retcode"] == 1


def test_retcode_exe_run_exception(salt_ssh_cli):
    """
    Verify salt-ssh passes through the retcode it receives
    when an exception is thrown. (Ref #50727)
    """
    ret = salt_ssh_cli.run("salttest.jinja_error")
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, dict)
    assert ret.data["stderr"].endswith("Exception: hehehe")
    assert ret.data["retcode"] == 1



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.extfs as extfs
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {extfs: {}}


# 'mkfs' function tests: 1


def test_mkfs():
    """
    Tests if a file system created on the specified device
    """
    mock = MagicMock()
    with patch.dict(extfs.__salt__, {"cmd.run": mock}):
        assert [] == extfs.mkfs("/dev/sda1", "ext4")


# 'tune' function tests: 1


def test_tune():
    """
    Tests if specified group was added
    """
    mock = MagicMock()
    with patch.dict(extfs.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.extfs.tune", MagicMock(return_value="")
    ):
        assert "" == extfs.tune("/dev/sda1")


# 'dump' function tests: 1


def test_dump():
    """
    Tests if specified group was added
    """
    mock = MagicMock()
    with patch.dict(extfs.__salt__, {"cmd.run": mock}):
        assert {"attributes": {}, "blocks": {}} == extfs.dump("/dev/sda1")


# 'attributes' function tests: 1


def test_attributes():
    """
    Tests if specified group was added
    """
    with patch(
        "salt.modules.extfs.dump",
        MagicMock(return_value={"attributes": {}, "blocks": {}}),
    ):
        assert {} == extfs.attributes("/dev/sda1")


# 'blocks' function tests: 1


def test_blocks():
    """
    Tests if specified group was added
    """
    with patch(
        "salt.modules.extfs.dump",
        MagicMock(return_value={"attributes": {}, "blocks": {}}),
    ):
        assert {} == extfs.blocks("/dev/sda1")



import pytest

import salt.modules.artifactory as artifactory
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {artifactory: {}}


def test_artifact_get_metadata():
    with patch(
        "salt.modules.artifactory._get_artifact_metadata_xml",
        MagicMock(
            return_value="""<?xml version="1.0" encoding="UTF-8"?>
            <metadata>
              <groupId>com.company.sampleapp.web-module</groupId>
              <artifactId>web</artifactId>
              <versioning>
                <latest>1.1_RC11</latest>
                <release>1.0.1</release>
                <versions>
                  <version>1.0_RC20</version>
                  <version>1.0_RC22</version>
                </versions>
                <lastUpdated>20140623120632</lastUpdated>
              </versioning>
            </metadata>
        """
        ),
    ):
        metadata = artifactory._get_artifact_metadata(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-releases",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            headers={},
        )
        assert metadata["latest_version"] == "1.1_RC11"


def test_snapshot_version_get_metadata():
    with patch(
        "salt.modules.artifactory._get_snapshot_version_metadata_xml",
        MagicMock(
            return_value="""<?xml version="1.0" encoding="UTF-8"?>
                <metadata>
                  <groupId>com.company.sampleapp.web-module</groupId>
                  <artifactId>web</artifactId>
                  <version>1.1_RC8-SNAPSHOT</version>
                  <versioning>
                    <snapshot>
                      <timestamp>20140418.150212</timestamp>
                      <buildNumber>1</buildNumber>
                    </snapshot>
                    <lastUpdated>20140623104055</lastUpdated>
                    <snapshotVersions>
                      <snapshotVersion>
                        <extension>pom</extension>
                        <value>1.1_RC8-20140418.150212-1</value>
                        <updated>20140418150212</updated>
                      </snapshotVersion>
                      <snapshotVersion>
                        <extension>war</extension>
                        <value>1.1_RC8-20140418.150212-1</value>
                        <updated>20140418150212</updated>
                      </snapshotVersion>
                    </snapshotVersions>
                  </versioning>
                </metadata>
            """
        ),
    ):
        metadata = artifactory._get_snapshot_version_metadata(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-releases",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            version="1.1_RC8-SNAPSHOT",
            headers={},
        )
        assert metadata["snapshot_versions"]["war"] == "1.1_RC8-20140418.150212-1"


def test_artifact_metadata_url():
    metadata_url = artifactory._get_artifact_metadata_url(
        artifactory_url="http://artifactory.example.com/artifactory",
        repository="libs-releases",
        group_id="com.company.sampleapp.web-module",
        artifact_id="web",
    )

    assert (
        metadata_url
        == "http://artifactory.example.com/artifactory/libs-releases/com/company/sampleapp/web-module/web/maven-metadata.xml"
    )


def test_snapshot_version_metadata_url():
    metadata_url = artifactory._get_snapshot_version_metadata_url(
        artifactory_url="http://artifactory.example.com/artifactory",
        repository="libs-snapshots",
        group_id="com.company.sampleapp.web-module",
        artifact_id="web",
        version="1.0_RC10-SNAPSHOT",
    )

    assert (
        metadata_url
        == "http://artifactory.example.com/artifactory/libs-snapshots/com/company/sampleapp/web-module/web/1.0_RC10-SNAPSHOT/maven-metadata.xml"
    )


def test_construct_url_for_released_version():
    artifact_url, file_name = artifactory._get_release_url(
        repository="libs-releases",
        group_id="com.company.sampleapp.web-module",
        artifact_id="web",
        packaging="war",
        artifactory_url="http://artifactory.example.com/artifactory",
        version="1.0_RC20",
    )

    assert (
        artifact_url
        == "http://artifactory.example.com/artifactory/libs-releases/com/company/sampleapp/web-module/web/1.0_RC20/web-1.0_RC20.war"
    )
    assert file_name == "web-1.0_RC20.war"


def test_construct_url_for_snapshot_version():
    with patch(
        "salt.modules.artifactory._get_snapshot_version_metadata",
        MagicMock(
            return_value={"snapshot_versions": {"war": "1.0_RC10-20131127.105838-2"}}
        ),
    ):

        artifact_url, file_name = artifactory._get_snapshot_url(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            version="1.0_RC10-SNAPSHOT",
            packaging="war",
            headers={},
        )

        assert (
            artifact_url
            == "http://artifactory.example.com/artifactory/libs-snapshots/com/company/sampleapp/web-module/web/1.0_RC10-SNAPSHOT/web-1.0_RC10-20131127.105838-2.war"
        )
        assert file_name == "web-1.0_RC10-20131127.105838-2.war"


def test_get_snapshot_url_with_classifier():
    with patch(
        "salt.modules.artifactory._get_snapshot_version_metadata_xml",
        MagicMock(
            return_value="""<?xml version="1.0" encoding="UTF-8"?>
                <metadata>
                  <groupId>com.company.sampleapp.web-module</groupId>
                  <artifactId>web</artifactId>
                  <version>1.1_RC8-SNAPSHOT</version>
                  <versioning>
                    <snapshot>
                      <timestamp>20140418.150212</timestamp>
                      <buildNumber>1</buildNumber>
                    </snapshot>
                    <lastUpdated>20140623104055</lastUpdated>
                    <snapshotVersions>
                      <snapshotVersion>
                        <extension>pom</extension>
                        <value>1.1_RC8-20140418.150212-1</value>
                        <updated>20140418150212</updated>
                      </snapshotVersion>
                      <snapshotVersion>
                        <classifier>test</classifier>
                        <extension>war</extension>
                        <value>1.1_RC8-20140418.150212-1</value>
                        <updated>20140418150212</updated>
                      </snapshotVersion>
                    </snapshotVersions>
                  </versioning>
                </metadata>
            """
        ),
    ):
        artifact_url, file_name = artifactory._get_snapshot_url(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            version="1.1_RC8-SNAPSHOT",
            packaging="war",
            classifier="test",
            headers={},
        )

        assert (
            artifact_url
            == "http://artifactory.example.com/artifactory/libs-snapshots/com/company/sampleapp/web-module/web/1.1_RC8-SNAPSHOT/web-1.1_RC8-20140418.150212-1-test.war"
        )


def test_get_snapshot_url_without_classifier():
    """
    test when classifier not set and packaging
    does not match snapshot_versions in the metadata.
    """
    with patch(
        "salt.modules.artifactory._get_snapshot_version_metadata_xml",
        MagicMock(
            return_value="""<?xml version="1.0" encoding="UTF-8"?>
                <metadata>
                  <groupId>com.company.sampleapp.web-module</groupId>
                  <artifactId>web</artifactId>
                  <version>1.1_RC8-SNAPSHOT</version>
                  <versioning>
                    <snapshot>
                      <timestamp>20140418.150212</timestamp>
                      <buildNumber>1</buildNumber>
                    </snapshot>
                    <lastUpdated>20140623104055</lastUpdated>
                    <snapshotVersions>
                      <snapshotVersion>
                        <extension>pom</extension>
                        <value>1.1_RC8-20140418.150212-1</value>
                        <updated>20140418150212</updated>
                      </snapshotVersion>
                      <snapshotVersion>
                        <classifier>test</classifier>
                        <extension>war</extension>
                        <value>1.1_RC8-20140418.150212-1</value>
                        <updated>20140418150212</updated>
                      </snapshotVersion>
                    </snapshotVersions>
                  </versioning>
                </metadata>
            """
        ),
    ):
        with pytest.raises(artifactory.ArtifactoryError):
            artifact_url, file_name = artifactory._get_snapshot_url(
                artifactory_url="http://artifactory.example.com/artifactory",
                repository="libs-snapshots",
                group_id="com.company.sampleapp.web-module",
                artifact_id="web",
                version="1.1_RC8-SNAPSHOT",
                packaging="war",
                headers={},
            )


def test_get_latest_snapshot_username_password():
    with patch(
        "salt.modules.artifactory._get_artifact_metadata",
        return_value={"latest_version": "1.1"},
    ), patch(
        "salt.modules.artifactory._get_snapshot_url",
        return_value=(
            "http://artifactory.example.com/artifactory/snapshot",
            "/path/to/file",
        ),
    ), patch(
        "salt.modules.artifactory.__save_artifact", return_value={}
    ) as save_artifact_mock:
        artifactory.get_latest_snapshot(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            packaging="war",
            username="user",
            password="password",
        )
        save_artifact_mock.assert_called_with(
            "http://artifactory.example.com/artifactory/snapshot",
            "/path/to/file",
            {"Authorization": "Basic dXNlcjpwYXNzd29yZA=="},
        )


def test_get_snapshot_username_password():
    with patch(
        "salt.modules.artifactory._get_snapshot_url",
        return_value=(
            "http://artifactory.example.com/artifactory/snapshot",
            "/path/to/file",
        ),
    ), patch(
        "salt.modules.artifactory.__save_artifact", return_value={}
    ) as save_artifact_mock:
        artifactory.get_snapshot(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            packaging="war",
            version="1.1",
            username="user",
            password="password",
        )
        save_artifact_mock.assert_called_with(
            "http://artifactory.example.com/artifactory/snapshot",
            "/path/to/file",
            {"Authorization": "Basic dXNlcjpwYXNzd29yZA=="},
        )


def test_get_latest_release_username_password():
    with patch(
        "salt.modules.artifactory.__find_latest_version",
        return_value="1.1",
    ), patch(
        "salt.modules.artifactory._get_release_url",
        return_value=(
            "http://artifactory.example.com/artifactory/release",
            "/path/to/file",
        ),
    ), patch(
        "salt.modules.artifactory.__save_artifact", return_value={}
    ) as save_artifact_mock:
        artifactory.get_latest_release(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            packaging="war",
            username="user",
            password="password",
        )
        save_artifact_mock.assert_called_with(
            "http://artifactory.example.com/artifactory/release",
            "/path/to/file",
            {"Authorization": "Basic dXNlcjpwYXNzd29yZA=="},
        )


def test_get_release_username_password():
    with patch(
        "salt.modules.artifactory._get_release_url",
        return_value=(
            "http://artifactory.example.com/artifactory/release",
            "/path/to/file",
        ),
    ), patch(
        "salt.modules.artifactory.__save_artifact", return_value={}
    ) as save_artifact_mock:
        artifactory.get_release(
            artifactory_url="http://artifactory.example.com/artifactory",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            packaging="war",
            version="1.1",
            username="user",
            password="password",
        )
        save_artifact_mock.assert_called_with(
            "http://artifactory.example.com/artifactory/release",
            "/path/to/file",
            {"Authorization": "Basic dXNlcjpwYXNzd29yZA=="},
        )


def test_save_artifact_file_exists_checksum_equal():
    artifact_url = "http://artifactory.example.com/artifactory/artifact"
    target_file = "/path/to/file"
    sum_str = "0123456789abcdef0123456789abcdef01234567"
    sum_bin = sum_str.encode()
    with patch("os.path.isfile", return_value=True), patch.dict(
        artifactory.__salt__, {"file.get_hash": MagicMock(return_value=sum_str)}
    ):
        with patch(
            "salt.modules.artifactory.__download",
            return_value=(True, sum_bin, None),
        ):
            result = getattr(artifactory, "__save_artifact")(
                artifact_url=artifact_url, target_file=target_file, headers={}
            )
            assert result == {
                "status": True,
                "changes": {},
                "target_file": target_file,
                "comment": (
                    "File {} already exists, checksum matches with Artifactory.\n"
                    "Checksum URL: {}.sha1".format(target_file, artifact_url)
                ),
            }
        with patch(
            "salt.modules.artifactory.__download",
            return_value=(True, sum_str, None),
        ):
            result = getattr(artifactory, "__save_artifact")(
                artifact_url=artifact_url, target_file=target_file, headers={}
            )
            assert result == {
                "status": True,
                "changes": {},
                "target_file": target_file,
                "comment": (
                    "File {} already exists, checksum matches with Artifactory.\n"
                    "Checksum URL: {}.sha1".format(target_file, artifact_url)
                ),
            }



import platform

import pytest

import salt.utils.files


@pytest.mark.parametrize(
    "options",
    [
        "",
        "  signed-by=/foo/bar  ",
        "  trusted=yes",
        "signed-by=/foo/bar arch=amd64,i386",
        "signed-by=foo/bar trusted=yes arch=amd64",
    ],
)
@pytest.mark.skipif(
    not any([x for x in ["ubuntu", "debian"] if x in platform.platform()]),
    reason="Test only for debian based platforms",
)
def test_adding_repo_file_options(states, tmp_path, options):
    """
    test adding a repo file using pkgrepo.managed
    and maintaining the user-supplied options
    """
    repo_file = str(tmp_path / "stable-binary.list")
    option = f"[{options}] " if options != "" else ""
    expected_option = f"[{options.strip()}] " if options != "" else ""
    repo_content = f"deb {option}http://www.deb-multimedia.org stable main"
    ret = states.pkgrepo.managed(name=repo_content, file=repo_file, clean_file=True)
    with salt.utils.files.fopen(repo_file, "r") as fp:
        file_content = fp.read()
        assert (
            file_content.strip()
            == f"deb {expected_option}http://www.deb-multimedia.org stable main"
        )



import time

import pytest


def _select_active_interface(interfaces):
    candidates = []
    for name, data in interfaces.items():
        if not data or name.lower().startswith("loopback"):
            continue
        flags = {flag.upper() for flag in data.get("flags", [])}
        is_up = data.get("up")
        if flags:
            if "UP" not in flags:
                continue
        elif is_up is False:
            continue
        candidates.append(name)
    if not candidates:
        pytest.skip("No active non-loopback interfaces available")
    return candidates[0]


def _wait_for_dns_state(
    salt_call_cli, interface, dns, *, present, timeout=30, interval=1
):
    deadline = time.time() + timeout
    last_servers = []
    while time.time() < deadline:
        ret = salt_call_cli.run("win_dns_client.get_dns_servers", interface=interface)
        assert ret.returncode == 0
        servers = ret.data or []
        last_servers = servers
        if present and dns in servers:
            return
        if not present and dns not in servers:
            return
        time.sleep(interval)

    expectation = "present" if present else "absent"
    pytest.fail(
        f"DNS server {dns} expected to be {expectation} on {interface}. "
        f"Last observed servers: {last_servers}"
    )


@pytest.mark.skip_unless_on_windows
@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
@pytest.mark.slow_test
def test_add_remove_dns(salt_call_cli):
    """Ensure a DNS server can be added and removed from a Windows interface."""

    ret = salt_call_cli.run("network.interfaces")
    assert ret.returncode == 0
    interfaces = ret.data or {}
    if not interfaces:
        pytest.skip("This test requires a network interface")

    interface = _select_active_interface(interfaces)
    dns = "8.8.8.8"

    ret = salt_call_cli.run("win_dns_client.get_dns_servers", interface=interface)
    assert ret.returncode == 0
    current_servers = ret.data or []
    index = len(current_servers) + 1

    ret = salt_call_cli.run("win_dns_client.add_dns", dns, interface, index=index)
    assert ret.returncode == 0
    assert ret.data is True

    _wait_for_dns_state(salt_call_cli, interface, dns, present=True)

    ret = salt_call_cli.run("win_dns_client.rm_dns", dns, interface=interface)
    assert ret.returncode == 0
    assert ret.data is True

    _wait_for_dns_state(salt_call_cli, interface, dns, present=False)



import json
import logging

import pytest

import salt.serializers.configparser
import salt.serializers.plist

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_serialize(file, tmp_path):
    """
    Test to ensure that file.serialize returns a data structure that's
    both serialized and formatted properly
    """
    path_test = tmp_path / "test_serialize"
    dataset = {
        "name": "naive",
        "description": "A basic test",
        "a_list": ["first_element", "second_element"],
        "finally": "the last item",
    }
    ret = file.serialize(
        name=str(path_test),
        dataset=dataset,
        formatter="json",
    )
    assert ret.result is True

    assert json.loads(path_test.read_text()) == dataset


def test_serializer_deserializer_opts(file, tmp_path):
    """
    Test the serializer_opts and deserializer_opts options
    """
    name = tmp_path / "testfile"

    data1 = {"foo": {"bar": "%(x)s"}}
    data2 = {"foo": {"abc": 123}}
    merged = {"foo": {"y": "not_used", "x": "baz", "abc": 123, "bar": "baz"}}

    ret = file.serialize(
        name=str(name),
        dataset=data1,
        formatter="configparser",
        deserializer_opts=[{"defaults": {"y": "not_used"}}],
    )
    assert ret.result is True

    # We should have warned about deserializer_opts being used when
    # merge_if_exists was not set to True.
    assert "warnings" in ret.filtered
    assert (
        "The 'deserializer_opts' option is ignored unless merge_if_exists is set to True."
        in ret.filtered["warnings"]
    )

    # Run with merge_if_exists, as well as serializer and deserializer opts
    # deserializer opts will be used for string interpolation of the %(x)s
    # that was written to the file with data1 (i.e. bar should become baz)
    ret = file.serialize(
        name=str(name),
        dataset=data2,
        formatter="configparser",
        merge_if_exists=True,
        serializer_opts=[{"defaults": {"y": "not_used"}}],
        deserializer_opts=[{"defaults": {"x": "baz"}}],
    )
    assert ret.result is True

    serialized_data = salt.serializers.configparser.deserialize(name.read_text())

    # If this test fails, this debug logging will help tell us how the
    # serialized data differs from what was serialized.
    log.debug("serialized_data = %r", serialized_data)
    log.debug("merged = %r", merged)
    # serializing with a default of 'y' will add y = not_used into foo
    assert serialized_data["foo"]["y"] == merged["foo"]["y"]
    # deserializing with default of x = baz will perform interpolation on %(x)s
    # and bar will then = baz
    assert serialized_data["foo"]["bar"] == merged["foo"]["bar"]


def test_serializer_plist_binary_file_open(file, tmp_path):
    """
    Test the serialization and deserialization of plists which should include
    the "rb" file open arguments change specifically for this formatter to handle
    binary plists.
    """
    name = tmp_path / "testfile"
    data1 = {"foo": {"bar": "%(x)s"}}
    data2 = {"foo": {"abc": 123}}
    merged = {"foo": {"abc": 123, "bar": "%(x)s"}}

    ret = file.serialize(
        name=str(name),
        dataset=data1,
        formatter="plist",
        serializer_opts=[{"fmt": "FMT_BINARY"}],
    )
    assert ret.result is True

    # Run with merge_if_exists so we test the deserializer.
    ret = file.serialize(
        name=str(name),
        dataset=data2,
        formatter="plist",
        merge_if_exists=True,
        serializer_opts=[{"fmt": "FMT_BINARY"}],
    )
    assert ret.result is True

    serialized_data = salt.serializers.plist.deserialize(name.read_bytes())

    # make sure our serialized data matches what we expect
    assert serialized_data["foo"] == merged["foo"]


def test_serializer_plist_file_open(file, tmp_path):
    """
    Test the serialization and deserialization of non binary plists with
    the new line concatenation.
    """
    name = tmp_path / "testfile"
    data1 = {"foo": {"bar": "%(x)s"}}
    data2 = {"foo": {"abc": 123}}
    merged = {"foo": {"abc": 123, "bar": "%(x)s"}}

    ret = file.serialize(name=str(name), dataset=data1, formatter="plist")
    assert ret.result is True

    # Run with merge_if_exists so we test the deserializer.
    ret = file.serialize(
        name=str(name),
        dataset=data2,
        formatter="plist",
        merge_if_exists=True,
    )
    assert ret.result is True

    serialized_data = salt.serializers.plist.deserialize(name.read_bytes())
    # make sure our serialized data matches what we expect
    assert serialized_data["foo"] == merged["foo"]



import os

import pytest
import yaml

# moto must be imported before boto3
try:
    import boto3
    from moto import mock_aws

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

import salt.fileserver.s3fs as s3fs
import salt.utils.s3

pytestmark = [
    pytest.mark.skipif(not HAS_BOTO, reason="Missing library moto or boto3"),
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def bucket():
    with mock_aws():
        yield "mybucket"


@pytest.fixture(scope="module")
def aws_creds():
    return {
        "aws_access_key_id": "testing",
        "aws_secret_access_key": "testing",
        "aws_session_token": "testing",
        "region_name": "us-east-1",
    }


@pytest.fixture(scope="function")
def configure_loader_modules(tmp_path, bucket):
    opts = {
        "cachedir": tmp_path,
        "s3.buckets": {"base": [bucket]},
        "s3.location": "us-east-1",
        "s3.s3_cache_expire": -1,
    }
    utils = {"s3.query": salt.utils.s3.query}

    yield {s3fs: {"__opts__": opts, "__utils__": utils}}


@pytest.fixture(scope="function")
def s3(bucket, aws_creds):
    conn = boto3.client("s3", **aws_creds)
    conn.create_bucket(Bucket=bucket)
    return conn


def make_keys(bucket, conn, keys):
    for key, data in keys.items():
        conn.put_object(
            Bucket=bucket,
            Key=key,
            Body=data["content"],
        )


def verify_cache(bucket, expected):
    for key, data in expected.items():
        correct_content = data["content"]
        cache_file = s3fs._get_cached_file_name(bucket, "base", key)
        assert os.path.exists(cache_file)

        if correct_content is None:
            continue

        with salt.utils.files.fopen(cache_file) as f:
            content = f.read()
            assert correct_content == content


@pytest.mark.skip_on_fips_enabled_platform
def test_update(bucket, s3):
    """Tests that files get downloaded from s3 to the local cache."""

    keys = {
        "top.sls": {"content": yaml.dump({"base": {"*": ["foo"]}})},
        "foo.sls": {"content": yaml.dump({"nginx": {"pkg.installed": []}})},
        "files/nginx.conf": {"content": "server {}"},
        "files/conf.d/foo.conf": {"content": "server {}"},
    }

    make_keys(bucket, s3, keys)
    s3fs.update()
    verify_cache(bucket, keys)

    # make a modification and update again - verify the change is retrieved
    keys["top.sls"]["content"] = yaml.dump({"base": {"*": ["foo", "bar"]}})
    make_keys(bucket, s3, keys)
    s3fs.update()
    verify_cache(bucket, keys)

    # verify that when files get deleted from s3, they also get deleted in
    # the local cache
    delete_file = "files/nginx.conf"
    del keys[delete_file]
    s3.delete_object(Bucket=bucket, Key=delete_file)

    s3fs.update()
    verify_cache(bucket, keys)

    cache_file = s3fs._get_cached_file_name(bucket, "base", delete_file)
    assert not os.path.exists(cache_file)

    # we want empty directories to get deleted from the local cache

    # after this one, `files` should still exist
    files_dir = os.path.dirname(cache_file)
    assert os.path.exists(files_dir)

    # but after the last file is deleted, the directory and any parents
    # should be deleted too
    delete_file = "files/conf.d/foo.conf"
    del keys[delete_file]
    s3.delete_object(Bucket=bucket, Key=delete_file)

    s3fs.update()
    verify_cache(bucket, keys)

    cache_file = s3fs._get_cached_file_name(bucket, "base", delete_file)
    assert not os.path.exists(cache_file)

    # after this, `files/conf.d` and `files` should be deleted
    conf_d_dir = os.path.dirname(cache_file)
    assert not os.path.exists(conf_d_dir)
    assert not os.path.exists(files_dir)


@pytest.mark.skip_on_fips_enabled_platform
def test_s3_hash(bucket, s3):
    """Verifies that s3fs hashes files correctly."""

    keys = {
        "top.sls": {"content": yaml.dump({"base": {"*": ["foo"]}})},
        "foo.sls": {"content": yaml.dump({"nginx": {"pkg.installed": []}})},
        "files/nginx.conf": {"content": "server {}"},
    }

    make_keys(bucket, s3, keys)
    s3fs.update()

    for key, item in keys.items():
        cached_file_path = s3fs._get_cached_file_name(bucket, "base", key)
        item["hash"] = salt.utils.hashutils.get_hash(
            cached_file_path, s3fs.S3_HASH_TYPE
        )
        item["cached_file_path"] = cached_file_path

    load = {"saltenv": "base"}
    fnd = {"bucket": bucket}

    for key, item in keys.items():
        fnd["path"] = item["cached_file_path"]
        actual_hash = s3fs.file_hash(load, fnd)
        assert s3fs.S3_HASH_TYPE == actual_hash["hash_type"]
        assert item["hash"] == actual_hash["hsum"]


@pytest.mark.skip_on_fips_enabled_platform
def test_cache_round_trip(bucket):
    metadata = {"foo": "bar"}
    cache_file = s3fs._get_buckets_cache_filename()
    s3fs._write_buckets_cache_file(metadata, cache_file)
    assert s3fs._read_buckets_cache_file(cache_file) == metadata


def test_ignore_pickle_load_exceptions():
    #  TODO: parameterized test with patched pickle.load that raises the
    #  various allowable exception from _read_buckets_cache_file
    pass



import urllib.parse

import pytest


@pytest.fixture
def client_config(client_config):
    client_config["rest_cherrypy"]["webhook_disable_auth"] = True
    return client_config


async def test_webhook_noauth(http_client):
    """
    Auth can be disabled for requests to the webhook URL

    See the above ``client_config`` fixture where we disable it
    """
    body = urllib.parse.urlencode({"foo": "Foo!"})
    response = await http_client.fetch("/hook", method="POST", body=body)
    assert response.code == 200



#
# Copyright 2015 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Philips HUE lamps module for proxy.

.. versionadded:: 2015.8.3
"""


import sys

__virtualname__ = "hue"
__proxyenabled__ = ["philips_hue"]


def _proxy():
    """
    Get proxy.
    """
    return __proxy__


def __virtual__():
    """
    Start the Philips HUE only for proxies.
    """
    if not _proxy():
        return False

    def _mkf(cmd_name, doc):
        """
        Nested function to help move proxy functions into sys.modules
        """

        def _cmd(*args, **kw):
            """
            Call commands in proxy
            """
            proxyfn = "philips_hue." + cmd_name
            return __proxy__[proxyfn](*args, **kw)

        return _cmd

    import salt.proxy.philips_hue as hue

    for method in dir(hue):
        if method.startswith("call_"):
            setattr(
                sys.modules[__name__],
                method[5:],
                _mkf(method, getattr(hue, method).__doc__),
            )
    del hue

    return _proxy() and __virtualname__ or False



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.sdb as sdb
from salt.exceptions import SaltInvocationError


@pytest.fixture
def configure_loader_modules():
    return {sdb: {}}


def test_get():
    """
    Test if it gets a value from a db, using a uri in the form of
    sdb://<profile>/<key>
    """
    assert sdb.get("sdb://salt/foo") == "sdb://salt/foo"


def test_get_strict_no_sdb_in_uri():
    """
    Test if SaltInvocationError exception will be raised if we
    don't start uri with sdb://
    """

    msg = 'SDB uri must start with "sdb://"'
    with pytest.raises(SaltInvocationError, match=msg) as cm:
        sdb.get("://salt/foo", strict=True)


def test_get_strict_no_profile():
    """
    Test if SaltInvocationError exception will be raised if we
    don't have a valid profile in the uri
    """

    msg = "SDB uri must have a profile name as a first part of the uri before the /"
    with pytest.raises(SaltInvocationError, match=msg) as cm:
        sdb.get("sdb://salt", strict=True)


def test_get_strict_no_profile_in_config():
    """
    Test if SaltInvocationError exception will be raised if we
    don't have expected profile in the minion config
    """

    msg = 'SDB profile "salt" wasnt found in the minion configuration'
    with pytest.raises(SaltInvocationError, match=msg) as cm:
        sdb.get("sdb://salt/foo", strict=True)


def test_set():
    """
    Test if it sets a value from a db, using a uri in the form of
    sdb://<profile>/<key>
    """
    assert not sdb.set_("sdb://mymemcached/foo", "bar")



"""
NAPALM Formula helpers
======================

.. versionadded:: 2019.2.0

This is an Execution Module providing helpers for various NAPALM formulas,
e.g., napalm-interfaces-formula, napalm-bgp-formula, napalm-ntp-formula etc.,
meant to provide various helper functions to make the templates more readable.
"""

import copy
import fnmatch
import logging

import salt.utils.dictupdate

# Import salt modules
import salt.utils.napalm
from salt.defaults import DEFAULT_TARGET_DELIM
from salt.utils.data import traverse_dict_and_list as _traverse_dict_and_list

__proxyenabled__ = ["*"]
__virtualname__ = "napalm_formula"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Available only on NAPALM Minions.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


def _container_path(model, key=None, container=None, delim=DEFAULT_TARGET_DELIM):
    """
    Generate all the possible paths within an OpenConfig-like object.
    This function returns a generator.
    """
    if not key:
        key = ""
    if not container:
        container = "config"
    for model_key, model_value in model.items():
        if key:
            key_depth = "{prev_key}{delim}{cur_key}".format(
                prev_key=key, delim=delim, cur_key=model_key
            )
        else:
            key_depth = model_key
        if model_key == container:
            yield key_depth
        else:
            yield from _container_path(
                model_value, key=key_depth, container=container, delim=delim
            )


def container_path(model, key=None, container=None, delim=DEFAULT_TARGET_DELIM):
    """
    Return the list of all the possible paths in a container, down to the
    ``config`` container.
    This function can be used to verify that the ``model`` is a Python object
    correctly structured and respecting the OpenConfig hierarchy.

    model
        The OpenConfig-structured object to inspect.

    delim: ``:``
        The key delimiter. In particular cases, it is indicated to use ``//``
        as ``:`` might be already used in various cases, e.g., IPv6 addresses,
        interface name (e.g., Juniper QFX series), etc.

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_formula.container_path "{'interfaces': {'interface': {'Ethernet1': {'config': {'name': 'Ethernet1'}}}}}"

    The example above would return a list with the following element:
    ``interfaces:interface:Ethernet1:config`` which is the only possible path
    in that hierarchy.

    Other output examples:

    .. code-block:: text

        - interfaces:interface:Ethernet1:config
        - interfaces:interface:Ethernet1:subinterfaces:subinterface:0:config
        - interfaces:interface:Ethernet2:config
    """
    return list(_container_path(model))


def setval(key, val, dict_=None, delim=DEFAULT_TARGET_DELIM):
    """
    Set a value under the dictionary hierarchy identified
    under the key. The target 'foo/bar/baz' returns the
    dictionary hierarchy {'foo': {'bar': {'baz': {}}}}.

    .. note::

        Currently this doesn't work with integers, i.e.
        cannot build lists dynamically.

    CLI Example:

    .. code-block:: bash

        salt '*' formula.setval foo:baz:bar True
    """
    if not dict_:
        dict_ = {}
    prev_hier = dict_
    dict_hier = key.split(delim)
    for each in dict_hier[:-1]:
        if each not in prev_hier:
            prev_hier[each] = {}
        prev_hier = prev_hier[each]
    prev_hier[dict_hier[-1]] = copy.deepcopy(val)
    return dict_


def traverse(data, key, default=None, delimiter=DEFAULT_TARGET_DELIM):
    """
    Traverse a dict or list using a colon-delimited (or otherwise delimited,
    using the ``delimiter`` param) target string. The target ``foo:bar:0`` will
    return ``data['foo']['bar'][0]`` if this value exists, and will otherwise
    return the dict in the default argument.
    Function will automatically determine the target type.
    The target ``foo:bar:0`` will return data['foo']['bar'][0] if data like
    ``{'foo':{'bar':['baz']}}`` , if data like ``{'foo':{'bar':{'0':'baz'}}}``
    then ``return data['foo']['bar']['0']``

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_formula.traverse "{'foo': {'bar': {'baz': True}}}" foo:baz:bar
    """
    return _traverse_dict_and_list(data, key, default=default, delimiter=delimiter)


def dictupdate(dest, upd, recursive_update=True, merge_lists=False):
    """
    Recursive version of the default dict.update

    Merges upd recursively into dest

    If recursive_update=False, will use the classic dict.update, or fall back
    on a manual merge (helpful for non-dict types like ``FunctionWrapper``).

    If ``merge_lists=True``, will aggregate list object types instead of replace.
    The list in ``upd`` is added to the list in ``dest``, so the resulting list
    is ``dest[key] + upd[key]``. This behaviour is only activated when
    ``recursive_update=True``. By default ``merge_lists=False``.
    """
    return salt.utils.dictupdate.update(
        dest, upd, recursive_update=recursive_update, merge_lists=merge_lists
    )


def defaults(model, defaults_, delim="//", flipped_merge=False):
    """
    Apply the defaults to a Python dictionary having the structure as described
    in the OpenConfig standards.

    model
        The OpenConfig model to apply the defaults to.

    defaults
        The dictionary of defaults. This argument must equally be structured
        with respect to the OpenConfig standards.

        For ease of use, the keys of these support glob matching, therefore
        we don't have to provide the defaults for each entity but only for
        the entity type. See an example below.

    delim: ``//``
        The key delimiter to use. Generally, ``//`` should cover all the possible
        cases, and you don't need to override this value.

    flipped_merge: ``False``
        Whether should merge the model into the defaults, or the defaults
        into the model. Default: ``False`` (merge the model into the defaults,
        i.e., any defaults would be overridden by the values from the ``model``).

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_formula.defaults "{'interfaces': {'interface': {'Ethernet1': {'config': {'name': 'Ethernet1'}}}}}" "{'interfaces': {'interface': {'*': {'config': {'enabled': True}}}}}"

    As one can notice in the example above, the ``*`` corresponds to the
    interface name, therefore, the defaults will be applied on all the
    interfaces.
    """
    merged = {}
    log.debug("Applying the defaults:")
    log.debug(defaults_)
    log.debug("openconfig like dictionary:")
    log.debug(model)
    for model_path in _container_path(model, delim=delim):
        for default_path in _container_path(defaults_, delim=delim):
            log.debug("Comparing %s to %s", model_path, default_path)
            if not fnmatch.fnmatch(model_path, default_path) or not len(
                model_path.split(delim)
            ) == len(default_path.split(delim)):
                continue
            log.debug("%s matches %s", model_path, default_path)
            # If there's a match, it will build the dictionary from the top
            devault_val = _traverse_dict_and_list(
                defaults_, default_path, delimiter=delim
            )
            merged = setval(model_path, devault_val, dict_=merged, delim=delim)
    log.debug("Complete default dictionary")
    log.debug(merged)
    log.debug("Merging with the model")
    log.debug(model)
    if flipped_merge:
        return salt.utils.dictupdate.update(model, merged)
    return salt.utils.dictupdate.update(merged, model)


def render_field(dictionary, field, prepend=None, append=None, quotes=False, **opts):
    """
    Render a field found under the ``field`` level of the hierarchy in the
    ``dictionary`` object.
    This is useful to render a field in a Jinja template without worrying that
    the hierarchy might not exist. For example if we do the following in Jinja:
    ``{{ interfaces.interface.Ethernet5.config.description }}`` for the
    following object:
    ``{'interfaces': {'interface': {'Ethernet1': {'config': {'enabled': True}}}}}``
    it would error, as the ``Ethernet5`` key does not exist.
    With this helper, we can skip this and avoid existence checks. This must be
    however used with care.

    dictionary
        The dictionary to traverse.

    field
        The key name or part to traverse in the ``dictionary``.

    prepend: ``None``
        The text to prepend in front of the text. Usually, we need to have the
        name of the field too when generating the configuration.

    append: ``None``
        Text to append at the end.

    quotes: ``False``
        Whether should wrap the text around quotes.

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_formula.render_field "{'enabled': True}" enabled
        # This would return the value of the ``enabled`` leaf key
        salt '*' napalm_formula.render_field "{'enabled': True}" description
        # This would not error

    Jinja usage example:

    .. code-block:: jinja

        {%- set config = {'enabled': True, 'description': 'Interface description'} %}
        {{ salt.napalm_formula.render_field(config, 'description', quotes=True) }}

    The example above would be rendered on Arista / Cisco as:

    .. code-block:: text

        description "Interface description"

    While on Junos (the semicolon is important to be added, otherwise the
    configuration won't be accepted by Junos):

    .. code-block:: text

        description "Interface description";
    """
    value = traverse(dictionary, field)
    if value is None:
        return ""
    if prepend is None:
        prepend = field.replace("_", "-")
    if append is None:
        if __grains__["os"] in ("junos",):
            append = ";"
        else:
            append = ""
    if quotes:
        value = f'"{value}"'
    return "{prepend} {value}{append}".format(
        prepend=prepend, value=value, append=append
    )


def render_fields(dictionary, *fields, **opts):
    """
    This function works similarly to
    :mod:`render_field <salt.modules.napalm_formula.render_field>` but for a
    list of fields from the same dictionary, rendering, indenting and
    distributing them on separate lines.

    dictionary
        The dictionary to traverse.

    fields
        A list of field names or paths in the dictionary.

    indent: ``0``
        The indentation to use, prepended to the rendered field.

    separator: ``\\n``
        The separator to use between fields.

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_formula.render_fields "{'mtu': 68, 'description': 'Interface description'}" mtu description

    Jinja usage example:

    .. code-block:: jinja

        {%- set config={'mtu': 68, 'description': 'Interface description'} %}
        {{ salt.napalm_formula.render_fields(config, 'mtu', 'description', quotes=True) }}

    The Jinja example above would generate the following configuration:

    .. code-block:: text

        mtu "68"
        description "Interface description"
    """
    results = []
    for field in fields:
        res = render_field(dictionary, field, **opts)
        if res:
            results.append(res)
    if "indent" not in opts:
        opts["indent"] = 0
    if "separator" not in opts:
        opts["separator"] = "\n{ind}".format(ind=" " * opts["indent"])
    return opts["separator"].join(results)



"""
    :codeauthor: Nicole Thomas (nicole@saltstack.com)
"""

import logging
from inspect import FullArgSpec

import pytest

import salt.states.module as module
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        module: {
            "__opts__": {"test": False},
            "__salt__": {CMD: MagicMock()},
            "__low__": {"__id__": "test"},
        }
    }


log = logging.getLogger(__name__)

CMD = "foo.bar"

STATE_APPLY_RET = {
    "module_|-test2_|-state.apply_|-run": {
        "comment": "Module function state.apply executed",
        "name": "state.apply",
        "start_time": "16:11:48.818932",
        "result": False,
        "duration": 179.439,
        "__run_num__": 0,
        "changes": {
            "ret": {
                "module_|-test3_|-state.apply_|-run": {
                    "comment": "Module function state.apply executed",
                    "name": "state.apply",
                    "start_time": "16:11:48.904796",
                    "result": True,
                    "duration": 89.522,
                    "__run_num__": 0,
                    "changes": {
                        "ret": {
                            "module_|-test4_|-cmd.run_|-run": {
                                "comment": "Module function cmd.run executed",
                                "name": "cmd.run",
                                "start_time": "16:11:48.988574",
                                "result": True,
                                "duration": 4.543,
                                "__run_num__": 0,
                                "changes": {"ret": "Wed Mar  7 16:11:48 CET 2018"},
                                "__id__": "test4",
                            }
                        }
                    },
                    "__id__": "test3",
                },
                "module_|-test3_fail_|-test3_fail_|-run": {
                    "comment": "Module function test3_fail is not available",
                    "name": "test3_fail",
                    "start_time": "16:11:48.994607",
                    "result": False,
                    "duration": 0.466,
                    "__run_num__": 1,
                    "changes": {},
                    "__id__": "test3_fail",
                },
            }
        },
        "__id__": "test2",
    }
}


def _mocked_func_named(
    name,
    names=(
        "Fred",
        "Swen",
    ),
):
    """
    Mocked function with named defaults.

    :param name:
    :param names:
    :return:
    """
    return {"name": name, "names": names}


def _mocked_func_args(*args):
    """
    Mocked function with args.

    :param args:
    :return:
    """
    assert args == ("foo", "bar")
    return {"args": args}


def _mocked_none_return(ret=None):
    """
    Mocked function returns None
    :return:
    """
    return ret


def test_run_module_not_available():
    """
    Tests the return of module.run state when the module function is not available.
    :return:
    """
    with patch.dict(module.__salt__, {}, clear=True), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: None})
    assert ret["comment"] == f"Unavailable function: {CMD}."
    assert ret["result"] is False


def test_run_module_not_available_testmode():
    """
    Tests the return of module.run state when the module function is not available
    when run with test=True
    :return:
    """
    with patch.dict(module.__salt__, {}, clear=True), patch.dict(
        module.__opts__, {"test": True, "use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: None})
    assert ret["comment"] == f"Unavailable function: {CMD}."
    assert ret["result"] is False


def test_run_module_noop():
    """
    Tests the return of module.run state when no module function is provided
    :return:
    """
    with patch.dict(module.__salt__, {}, clear=True), patch.dict(
        module.__opts__, {"test": True, "use_superseded": ["module.run"]}
    ):
        ret = module.run()
    assert ret["comment"] == "No function provided."
    assert ret["result"] is False


def test_module_run_hidden_varargs():
    """
    Tests the return of module.run state when hidden varargs are used with
    wrong type.
    """
    spec = FullArgSpec(
        args=[],
        varargs="names",
        varkw=None,
        defaults=None,
        kwonlyargs="kwargs",
        kwonlydefaults=None,
        annotations=None,
    )
    with patch("salt.utils.args.get_function_argspec", MagicMock(return_value=spec)):
        ret = module._legacy_run(CMD, m_names="anyname")
    assert ret["comment"] == "'names' must be a list."


def test_run_testmode():
    """
    Tests the return of the module.run state when test=True is passed.
    :return:
    """
    with patch.dict(module.__opts__, {"test": True, "use_superseded": ["module.run"]}):
        ret = module.run(**{CMD: None})
    assert ret["comment"] == f"Function {CMD} to be executed."
    assert ret["result"] is None


def test_run_missing_arg():
    """
    Tests the return of module.run state when arguments are missing
    :return:
    """
    with patch.dict(module.__salt__, {CMD: _mocked_func_named}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: None})
    assert ret["comment"] == f"'{CMD}' failed: Missing arguments: name"


def test_run_correct_arg():
    """
    Tests the return of module.run state when arguments are correct
    :return:
    """
    with patch.dict(module.__salt__, {CMD: _mocked_func_named}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: ["Fred"]})
    assert ret["comment"] == f"{CMD}: Success"
    assert ret["result"] is True


def test_run_state_apply_result_false():
    """
    Tests the 'result' of module.run that calls state.apply execution module
    :return:
    """
    with patch.dict(
        module.__salt__, {"state.apply": MagicMock(return_value=STATE_APPLY_RET)}
    ), patch.dict(module.__opts__, {"use_deprecated": ["module.run"]}):
        ret = module.run(**{"name": "state.apply", "mods": "test2"})
    assert ret["result"] is False


def test_run_service_status_dead():
    """
    Tests the 'result' of module.run that calls service.status when
    service is dead or does not exist
    """
    func = "service.status"
    with patch.dict(module.__salt__, {func: MagicMock(return_value=False)}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(
            **{
                "name": "test_service_state",
                "service.status": {"name": "doesnotexist"},
            }
        )
        assert ret == {
            "name": ["service.status"],
            "changes": {},
            "comment": "'service.status': False",
            "result": False,
        }


def test_run_unexpected_keywords():
    with patch.dict(module.__salt__, {CMD: _mocked_func_args}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: [{"foo": "bar"}]})
        module_function = module.__salt__[CMD].__name__
        assert ret[
            "comment"
        ] == "'{}' failed: {}() got an unexpected keyword argument 'foo'".format(
            CMD, module_function
        )
        assert ret["result"] is False


def test_run_args():
    """
    Test unnamed args.
    :return:
    """
    with patch.dict(module.__salt__, {CMD: _mocked_func_args}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: ["foo", "bar"]})
    assert ret["result"] is True
    assert ret["changes"] == {CMD: {"args": ("foo", "bar")}}


def test_run_42270():
    """
    Test example provided in issue 42270
    """

    def test_func(arg1, arg2, **kwargs):
        return {"args": [arg1, arg2], "kwargs": kwargs or {}}

    with patch.dict(module.__salt__, {CMD: test_func}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: ["bla", {"example": "bla"}]})
    assert ret["result"] is False
    assert ret["comment"] == f"'{CMD}' failed: Missing arguments: arg2"


def test_run_42270_kwargs_to_args():
    """
    Test module.run filling in args with kwargs with the same name.
    """

    def test_func(arg1, arg2, arg3, *args, **kwargs):
        return {"args": [arg1, arg2, arg3] + list(args), "kwargs": kwargs}

    with patch.dict(module.__salt__, {CMD: test_func}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: ["foo", "bar", {"arg3": "baz"}, {"foo": "bar"}]})

    assert ret["result"] is True
    assert ret["changes"] == {
        CMD: {"args": ["foo", "bar", "baz"], "kwargs": {"foo": "bar"}}
    }


def test_run_none_return():
    """
    Test handling of a broken function that returns None.
    :return:
    """
    with patch.dict(module.__salt__, {CMD: _mocked_none_return}), patch.dict(
        module.__opts__, {"use_superseded": ["module.run"]}
    ):
        ret = module.run(**{CMD: None})
    assert ret["result"] is True
    assert ret["changes"] == {CMD: None}


def test_run_typed_return():
    """
    Test handling of a broken function that returns any type.
    :return:
    """
    for val in [
        1,
        0,
        "a",
        "",
        (
            1,
            2,
        ),
        (),
        [1, 2],
        [],
        {"a": "b"},
        {},
        True,
        False,
    ]:
        with patch.dict(module.__salt__, {CMD: _mocked_none_return}), patch.dict(
            module.__opts__, {"use_superseded": ["module.run"]}
        ):
            log.debug("test_run_typed_return: trying %s", val)
            ret = module.run(**{CMD: [{"ret": val}]})
        if val is False:
            assert ret["result"] is False
            assert ret["comment"] == "'foo.bar': False"
        else:
            assert ret["result"] is True


def test_run_batch_call():
    """
    Test batch call
    :return:
    """
    with patch.dict(module.__opts__, {"use_superseded": ["module.run"]}), patch.dict(
        module.__salt__,
        {
            "first.one": _mocked_none_return,
            "second.one": _mocked_none_return,
            "third.one": _mocked_none_return,
        },
        clear=True,
    ):
        for f_name in module.__salt__:
            log.debug("test_run_batch_call: trying %s", f_name)
            ret = module.run(**{f_name: None})
            assert ret["result"] is True


def test_module_run_module_not_available():
    """
    Tests the return of module.run state when the module function
    name isn't available
    """
    with patch.dict(module.__salt__, {}, clear=True):
        ret = module._legacy_run(CMD)
    assert ret["result"] is False
    assert ret["comment"] == f"Module function {CMD} is not available"


def test_module_run_test_true():
    """
    Tests the return of module.run state when test=True is passed in
    """
    with patch.dict(module.__opts__, {"test": True}):
        ret = module._legacy_run(CMD)
    assert ret["comment"] == f"Module function {CMD} is set to execute"


def test_module_run_missing_arg():
    """
    Tests the return of module.run state when arguments are missing
    """
    spec = FullArgSpec(
        args=["hello", "world"],
        varargs=None,
        varkw=None,
        defaults=False,
        kwonlyargs=None,
        kwonlydefaults=None,
        annotations=None,
    )

    with patch("salt.utils.args.get_function_argspec", MagicMock(return_value=spec)):
        ret = module._legacy_run(CMD)
    assert "The following arguments are missing:" in ret["comment"]
    assert "world" in ret["comment"]
    assert "hello" in ret["comment"]


def test_call_function_named_args():
    """
    Test _call_function routine when params are named. Their position ordering should not matter.

    :return:
    """
    with patch.dict(
        module.__salt__,
        {"testfunc": lambda a, b, c, *args, **kwargs: (a, b, c, args, kwargs)},
        clear=True,
    ):
        assert module._call_function(
            "testfunc", func_args=[{"a": 1}, {"b": 2}, {"c": 3}]
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_args=[{"c": 3}, {"a": 1}, {"b": 2}]
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_kwargs={"a": 1, "b": 2, "c": 3}
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_kwargs={"c": 3, "a": 1, "b": 2}
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_kwargs={"a": 1, "c": 3, "b": 2}
        ) == (1, 2, 3, (), {})

    with patch.dict(
        module.__salt__,
        {"testfunc": lambda c, a, b, *args, **kwargs: (a, b, c, args, kwargs)},
        clear=True,
    ):
        assert module._call_function(
            "testfunc", func_args=[{"a": 1}, {"b": 2}, {"c": 3}]
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_args=[{"c": 3}, {"a": 1}, {"b": 2}]
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_kwargs={"a": 1, "b": 2, "c": 3}
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_kwargs={"c": 3, "a": 1, "b": 2}
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_kwargs={"b": 2, "a": 1, "c": 3}
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_kwargs={"a": 1, "c": 3, "b": 2}
        ) == (1, 2, 3, (), {})


def test_call_function_ordered_args():
    """
    Test _call_function routine when params are not named. Their position should matter.

    :return:
    """
    with patch.dict(
        module.__salt__,
        {"testfunc": lambda a, b, c, *args, **kwargs: (a, b, c, args, kwargs)},
        clear=True,
    ):
        assert module._call_function("testfunc", func_args=[1, 2, 3]) == (
            1,
            2,
            3,
            (),
            {},
        )
        assert module._call_function("testfunc", func_args=[3, 1, 2]) == (
            3,
            1,
            2,
            (),
            {},
        )


def test_call_function_ordered_and_named_args():
    """
    Test _call_function routine when params are not named. Their position should matter.

    :return:
    """
    with patch.dict(
        module.__salt__,
        {"testfunc": lambda a, b, c, *args, **kwargs: (a, b, c, args, kwargs)},
        clear=True,
    ):
        assert module._call_function(
            "testfunc", func_args=[1], func_kwargs={"b": 2, "c": 3}
        ) == (1, 2, 3, (), {})
        assert module._call_function(
            "testfunc", func_args=[1, 2], func_kwargs={"c": 3}
        ) == (1, 2, 3, (), {})


def test_get_result():
    """
    Test _get_result routine when function returned result or retcode in it's results or in changes:

    :return:
    """

    # List the tests with it's expected result:
    tests = (
        ({"result": False}, False),
        ({"retcode": "-1"}, False),
        ({"result": True}, True),
        ({"retcode": 0}, True),
        ({"retcode": -1, "result": True}, False),
        (True, True),
        (False, False),
        ({"changes": {"result": False}}, False),
        ({"changes": {"retcode": 1}}, False),
        ({"changes": {"result": True}}, True),
        ({"changes": {"retcode": 0}}, True),
        ({"changes": {"result": False, "retcode": 0}}, False),
        ({"changes": {"result": True, "retcode": 1}}, False),
        ({"changes": {"result": False, "retcode": 1}}, False),
        ({"changes": {"result": True, "retcode": 0}}, True),
    )

    for test in tests:
        ret = module._get_result(test[0])
        assert ret is test[1]
